//
//  BEBBCWindowController.m
//  BeebEmX
//
//  Created by Steven Bagley on 30/04/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#import <OpenGL/OpenGL.h>
#import <OpenGL/gl.h>
#import <OpenGL/glu.h>
#import "BEBBCWindowController.h"
#import "BEBBCWindow.h"
#include "BeebFrame.h"

#include "FrameQueue.h"

static CVReturn MyDisplayLinkCallback(CVDisplayLinkRef displayLink, const CVTimeStamp *inNow, const CVTimeStamp *inOutputTime, CVOptionFlags flagsIn, CVOptionFlags *flagsOut,
                                      void *displayLinkContext)
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    
    [(BEBBCWindowController *)displayLinkContext render];
    
    [pool drain];
    return kCVReturnSuccess;
}

@interface BEBBCWindowController ()

@end

@implementation BEBBCWindowController


-(BEBBCWindowController*)initWithOpenGL:(BOOL)useGL
{
    BEBBCWindow *myWindow;
    
    myWindow = [[BEBBCWindow alloc] initWithContentRect:NSMakeRect(100.0,100.0, 768.0, 576.0) styleMask:(NSTitledWindowMask|NSClosableWindowMask|NSMiniaturizableWindowMask|NSResizableWindowMask) backing:NSBackingStoreBuffered defer:YES];
    [myWindow setTitle:@"BBC Micro"];
    
    
    self = [super initWithWindow:myWindow];
    
    if(self)
    {
        _useGL = useGL;
        
        if(_useGL)
        {
            NSOpenGLPixelFormatAttribute pixelAttribs[16];
            NSOpenGLPixelFormatAttribute *p = pixelAttribs;
            
            *p++ = NSOpenGLPFADoubleBuffer;
            *p++ = NSOpenGLPFAAccelerated;
            *p++ = NSOpenGLPFANoRecovery;
            *p++ = NSOpenGLPFAColorSize;
            *p++ = 32;
            *p++ = NSOpenGLPFADepthSize;
            *p++ = 16;
            *p++ = NSOpenGLPFASupersample;
            *p++ = NSOpenGLPFASampleBuffers;
            *p++ = 1;
            *p++ = NSOpenGLPFASamples;
            *p++ = 4;
            *p++ = 0;
            
            NSOpenGLPixelFormat *pixelFormat = [[NSOpenGLPixelFormat alloc] initWithAttributes:pixelAttribs ];
            
            _bbcGLView = [[NSOpenGLView alloc] initWithFrame:[self.window contentRectForFrameRect:self.window.frame] pixelFormat:pixelFormat];
            self.window.contentView = _bbcGLView;
            [pixelFormat release];
            [self initGL:[_bbcGLView openGLContext]];
        }
        else
        {
            _bbcView = [[BEBBCView alloc] initWithFrame:NSMakeRect(0,0,768.0,576.0)];
            self.window.contentView = _bbcView;
        }
        
        _emVideoLink = (EMVIDEOLINK*)malloc(sizeof(EMVIDEOLINK));
        _emVideoLink->frameQueue = new std::queue<CBeebFrame*>;
        if(0 != pthread_mutex_init(&_emVideoLink->queueMutex, NULL))
        {
            NSLog(@"Failed to create frame queue mutex");
        }
  
    }
    else
    {
        [myWindow release];
    }
    
    return self;
}
-(void)initGL:(NSOpenGLContext*)gl
{
    GLint swapInterval = 1;

    
    [gl makeCurrentContext];
    [gl setValues:&swapInterval forParameter:NSOpenGLCPSwapInterval];
    
    glShadeModel(GL_SMOOTH);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    glClearDepth(1.0f);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    
    
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA_SATURATE, GL_ONE);
    glEnable(GL_POLYGON_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    
    glEnable(GL_MULTISAMPLE_ARB);
    
    glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);    glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glShadeModel(GL_SMOOTH);
///    glEnable(GL_LIGHTING);
    
    [_bbcGLView update];

    glViewport(0.0f,0.0f, (GLfloat)768.0, (GLfloat)576.0);

    /* Now create a texture to map onto the shape */
    
    glGenTextures(1, &_texture);
    glBindTexture(GL_TEXTURE_2D, _texture);

    
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 768.0f,
                 576.0f, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                 NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                    GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                    GL_LINEAR);

    //    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glGenBuffers(1, &_textureBuffer);
//GL_DYNAMIC_DRAW

    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, _textureBuffer);
    glBufferData(GL_PIXEL_UNPACK_BUFFER, 768*576*4, NULL, GL_STREAM_DRAW);

#if 0
    unsigned char *p = (unsigned char *)glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);
    
    for(int i = 0; i < 768 *288; i++)
    {
        *p++ = i;
        *p++ = i;
        *p++ = i;
        *p++ = i;
    }
    glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);
#endif
    
}

-(void)startEmulator:(id)sender
{
    CVTime refresh;
    
    
    CVDisplayLinkCreateWithActiveCGDisplays(&_dlRef);
    CVDisplayLinkSetOutputCallback(_dlRef, &MyDisplayLinkCallback, self);
    refresh = CVDisplayLinkGetNominalOutputVideoRefreshPeriod(_dlRef);
    _refreshPeriod = (double)refresh.timeValue / (double)refresh.timeScale;
    NSLog(@"Starting Thread");
    
    
    
    _bbcEmulator = [[BEEmulator alloc] initWithVideoLink:_emVideoLink];
    [_bbcEmulator start];
    CVDisplayLinkStart(_dlRef);
    [_bbcEmulator runForPeriod:_refreshPeriod];
    
}

-(void)renderGLFrame:(CBeebFrame*)newFrame
{
    static unsigned int beebPalette[] =
    {
        /* ABGR */
        0x00000000,  /* Black */
        0xFF0000FF,  /* Red */
        0xFF00FF00,  /* Green */
        0xFF00FFFF,  /* Yellow */
        0xFFFF0000,  /* Blue */
        0xFFFF00FF,  /* Magenta */
        0xFFFFFF00,  /* Cyan */
        0xFFFFFFFF   /* White */
    };
    
    [_bbcGLView lockFocusIfCanDraw];
    [[_bbcGLView openGLContext] makeCurrentContext];
    
    
    /* Set texture */
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, _texture);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, _textureBuffer);

    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 768, 576, GL_RGBA, GL_UNSIGNED_BYTE, 0);



    glBufferData(GL_PIXEL_UNPACK_BUFFER, 768*576*4, NULL, GL_STREAM_DRAW);
    
 
    unsigned int *p = (unsigned int *)glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);;
    unsigned char *q = newFrame->Data();
    int i;
    
    for(i = 0; i < 768*576; i++)
    {
        *p++ = beebPalette[*q++];
        
    }
    

    glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode( GL_PROJECTION );   // Select the projection matrix
    glLoadIdentity();                // and reset it
    // Calculate the aspect ratio of the view
    /* Angle of view needs to be calculated here
     *
     * If we are at 10.0f units back from the screen at 0.0f units,
     * and the screen goes from -2.0f to 2.0f then tan T = 2/10,
     */
    
//    NSLog(@"Field of View is %f", atan(2.0/10.0)*2.0 / M_PI*180.0);
    gluPerspective( atan(1.5/10.0)*2.0 / M_PI*180.0, 768.0 / 576.0, 0.1f, 15.0f );
    
    glMatrixMode( GL_MODELVIEW );    // Select the modelview matrix
    glLoadIdentity();                // and reset it
    
    gluLookAt(0.0f, 0.0f, 10.0f,   -0.0f, 0.0f, 0, 0.0f,1.0f, 0.0f );
    
    // Really nice perspective calculations
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
//    glColor4f(1.0f,1.0f,1.0f,1.0f);
    GLfloat global_ambient[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

    
    
 

    
    glColor4f(1.0f,1.0f,1.0f,1.0f);
    glNormal3f(0.0f, 0.0f, 1.0f);
    glBegin(GL_QUADS);
        glTexCoord2f(0.0f, 0.00f);      glVertex3f(-2.0f, 1.5f, 0.0f);
        glTexCoord2f(0.0f, 1.00f);      glVertex3f(-2.0f, -1.5f,0.0f);
        glTexCoord2f(1.0f, 1.00f);      glVertex3f(2.0f, -1.5f, 0.0f);
        glTexCoord2f(1.0f, 0.00f);      glVertex3f(2.0f, 1.5f, -0.0f);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, 0);

    
    
    glFlush();
    [[_bbcGLView openGLContext] flushBuffer];

    [_bbcGLView unlockFocus];

}

-(void)render
{
    CBeebFrame *frame = NULL;
    //    [_bbcController update];
    
    /* get frame */
    pthread_mutex_lock(&_emVideoLink->queueMutex);
    
    if(!_emVideoLink->frameQueue->empty())
    {
        frame =_emVideoLink->frameQueue->front();
        _emVideoLink->frameQueue->pop();
    }
    pthread_mutex_unlock(&_emVideoLink->queueMutex);
    
    [_bbcEmulator runForPeriod:_refreshPeriod];
    
    /* Display Frame */
    if(frame)
    {
        if(_useGL)
        {
            [self renderGLFrame:frame];
        }
        else
        {
            [self updateWithFrame:frame];
        }
        delete frame;
    }
}

-(void)flagsChanged:(NSEvent*)theEvent
{
    [_bbcEmulator flagsChanged:(unsigned int)[theEvent modifierFlags]];
}

-(void)updateWithFrame:(CBeebFrame*)newFrame
{
    [_bbcView fillWithFrame:newFrame];
    [_bbcView setNeedsDisplay:YES];
}

-(void)keyDown:(NSEvent *)theEvent
{
//    NSLog(@"Key down: %d", [theEvent keyCode]);
    [_bbcEmulator keyDown:[theEvent keyCode]];
    
}

- (void)keyUp:(NSEvent *)theEvent
{
//    NSLog(@"Key up: %d", [theEvent keyCode]);
    [_bbcEmulator keyUp:[theEvent keyCode]];
}

-(IBAction)loadDiscImage:(id)sender
{
    NSOpenPanel *panel;
    NSLog(@"Drop sheet");
    
    panel = [NSOpenPanel openPanel];
    [panel setCanChooseFiles:YES];
    [panel setCanChooseDirectories:NO];
    [panel setTitle:@"Load Disk"];
    [panel setAllowedFileTypes:[NSArray arrayWithObjects:@"ssd", @"sdd", nil]];
    [panel setAllowsOtherFileTypes:YES];
    
    [panel beginSheetModalForWindow:[self window] completionHandler:^(NSInteger result)
     {
         if (result == NSFileHandlingPanelOKButton)
         {
             NSURL*  url = [[panel URLs] objectAtIndex:0];
             
             [_bbcEmulator loadDisk:url];
             
             // Open  the document.
         }
     }];
}



@end
