//
//  BEEmulator.m
//  BeebEmX
//
//  Created by Steven Bagley on 20/04/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#import <Cocoa/Cocoa.h>
#import "BEEmulator.h"
#include "BeebEm.h"
#include "BeebMemory.h"
#include "FrameQueue.h"

static int *keyTable = NULL;


@implementation BEEmulator


-(BEEmulator*)initWithVideoLink:(EMVIDEOLINK *)vlink
{
    self = [super init];
    
    if(self)
    {
        _condition = [[NSCondition alloc] init];
        _runToTime = 0.0;
        _beeb = NULL;
        _emulatorLock = [[NSLock alloc] init];
        
        if(!keyTable)
        {
            NSString *path;
            
            keyTable = (int*)calloc(1,sizeof(int)*2*256);
            path = [[NSBundle mainBundle] pathForResource:@"KeyMapping" ofType:@"csv"];
            
            if(path)
            {
                int row, col, keycode;
                char line[128];
                FILE *fp = fopen([path UTF8String], "r");
                
                while(!feof(fp))
                {
                    fgets(line, 128, fp);
                    if(0 == strcmp(line, "\n"))
                    {
                        break;
                    }
                    
                    row = col = 0;
                    sscanf(line, "%x,%d,%d", &keycode, &row, &col);
                    NSLog(@"Keycode %d is %d %d", keycode, row, col);
                    keyTable[keycode * 2 + 0] = row;
                    keyTable[keycode * 2 + 1] = col;
                    
                }
                
                
                fclose(fp);
            }
            
            _previousModifiers = 0;
        }
        _vlink = vlink;
        
    }
    
    
    return self;
}


-(void)main
{
    NSString *romPath;
    NSData *rom;
    
    _beeb = new CBeebEm();
    
    _beeb->Initialise(_vlink);
//    _beeb->InitialiseMemory();
    
    /* Load ROMs */
    romPath = [[NSBundle mainBundle] pathForResource:@"os12" ofType:@"rom"];
    rom = [NSData dataWithContentsOfFile:romPath];
  
    _beeb->Memory()->LoadOS(new CBeebMemoryRom([rom bytes], (unsigned int)[rom length]));
    
//    _beeb->LoadOSRom([rom bytes], (unsigned int)[rom length]);
    
    romPath = [[NSBundle mainBundle] pathForResource:@"basic2" ofType:@"rom"];
    rom = [NSData dataWithContentsOfFile:romPath];
    _beeb->Memory()->LoadSWR(15, new CBeebMemoryRom([rom bytes], (unsigned int)[rom length]));

  //  romPath = [[NSBundle mainBundle] pathForResource:@"ADFS130" ofType:@"rom"];
 //   rom = [NSData dataWithContentsOfFile:romPath];
//    _beeb->Memory()->LoadSWR(14, new CBeebMemoryRom([rom bytes], (unsigned int)[rom length]));

    romPath = [[NSBundle mainBundle] pathForResource:@"DFS-2.26" ofType:@"rom"];
    rom = [NSData dataWithContentsOfFile:romPath];
    _beeb->Memory()->LoadSWR(13, new CBeebMemoryRom([rom bytes], (unsigned int)[rom length]));
    
    romPath = [[NSBundle mainBundle] pathForResource:@"DoctorWhoAndTheMinesOfTerror" ofType:@"rom"];
    rom = [NSData dataWithContentsOfFile:romPath];
    _beeb->Memory()->LoadSWR(12, new CBeebMemoryRom([rom bytes], (unsigned int)[rom length]));
    
    romPath = [[NSBundle mainBundle] pathForResource:@"Teletext" ofType:@"fnt"];
    rom = [NSData dataWithContentsOfFile:romPath];
    _beeb->LoadMode7Font((const unsigned char*)[rom bytes]);
    
    _beeb->Set1770DFS();
    
    _beeb->Reset();
//    _beeb->KeyDown(0,0); /* Shift */
    
    romPath = [[NSBundle mainBundle] pathForResource:@"Painter" ofType:@"ssd"];
    _beeb->LoadDisc((char *)[romPath UTF8String], 0,0);
    
    while(1)
    {
        [_condition lock];
        while((_beeb->CyclesExecuted() / 2000000.0) >= _runToTime)
        {
 //           NSLog(@"%d %f %f",(int)clock(), (beeb->CyclesExecuted() / 2000000.0), _runToTime);
            CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0, false);
            [_condition wait];
        }
        
        
        [_emulatorLock lock];
        while((_beeb->CyclesExecuted() / 2000000.0) < _runToTime)
        {
//            NSLog(@"Tick %f %f",(beeb->CyclesExecuted() / 2000000.0), _runToTime);

            _beeb->Tick();
            CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0, false);
        }
        [_emulatorLock unlock];
        [_condition unlock];
    }
    
    
}

-(void)loadDisk:(NSURL*)url
{
    NSString *path = [url path];
    
    _beeb->EjectDisc(0);
    _beeb->LoadDisc((char*)[path UTF8String], 0, 0);
}

-(void)runToTime:(double)time
{
    [_condition lock];
    _runToTime = time;
    [_condition signal];
    [_condition unlock];
}

-(void)runForPeriod:(double)time
{
    [_condition lock];
    _runToTime += time;
    [_condition signal];
    [_condition unlock];
    
}

-(void)keyDown:(unsigned short)keyCode
{
    [_emulatorLock lock];
    _beeb->KeyDown(keyTable[keyCode * 2],keyTable[keyCode * 2 + 1]);
    [_emulatorLock unlock];
}

-(void)keyUp:(unsigned short)keyCode
{
    [_emulatorLock lock];
    _beeb->KeyUp(keyTable[keyCode * 2],keyTable[keyCode * 2 + 1]);
    [_emulatorLock unlock];
}

-(void)flagsChanged:(unsigned int)modifiers
{
    unsigned int xorFlags;

    /* modifiers */
    xorFlags = modifiers ^ _previousModifiers;
    _previousModifiers = modifiers;
    
    if(xorFlags)
    {
        [_emulatorLock lock];
        
        if(xorFlags & NSShiftKeyMask)
        {
            if(modifiers & NSShiftKeyMask)
            {
                _beeb->KeyDown(0,0); /* Shift */
            }
            else
            {
                _beeb->KeyUp(0,0);
            }
            
        }
        
        [_emulatorLock unlock];
    }

}

@end
