//
//  BeebFrame.h
//  BeebEmX
//
//  Created by Steven Bagley on 03/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#ifndef __BeebEmX__BeebFrame__
#define __BeebEmX__BeebFrame__


#include "video.h"

class CBeebFrame
{
public:
    CBeebFrame()
    {
        m_screen = (unsigned char *)malloc(kFrameWidth * kFrameHeight);
    }
    
    ~CBeebFrame()
    {
        free(m_screen);
    }

    static const int kFrameWidth = 768;
    static const int kFrameHeight = 576;

    inline void doHorizLine(int Col, int y, int sx, int width)
    {
        int x;
        unsigned char *p;
        
        if (TeletextEnabled)
        {
            y /= TeletextStyle;
        }
        
//        int d = 4 * ((y * kFrameWidth) + sx + ScreenAdjust + (TeletextEnabled ? 36 : 0));
        int d;
        
        d = y * kFrameWidth;
        d += sx;
        d += ScreenAdjust;
        d += (TeletextEnabled ? 36 : 0);
        
        if((d + width) > (500 * kFrameWidth))
        {
            return;
        }
        if(d < 0)
        {
            return;
        }
        
        p = (unsigned char*)(m_screen + d);
        
        for(x = 0; x < width; x++)
        {
            *p++ = Col;
        }
    };
    
    void doInvHorizLine(int Col, int y, int sx, int width)
    {
        unsigned char *p;

        if(TeletextEnabled)
        {
            y /= TeletextStyle;
        }
        
//        int d = (y*800)+sx+ScreenAdjust+(TeletextEnabled?36:0);

        int d;
  
        d  = y * kFrameWidth;
        d += sx;
        d += ScreenAdjust;
        d += (TeletextEnabled ? 36 : 0);

        
        char *vaddr;
        
        if((d + width) > (500 * kFrameWidth))
        {
            return;
        }
        
        if(d<0)
        {
            return;
        }
        
        p = (unsigned char *)(m_screen + d);
        for(int n = 0; n < width; n++)
        {
            *p++ ^= Col;
        }
    };
    
    
    void ClearScanLine(int y)
    {
        int d = y * kFrameWidth + ScreenAdjust;
        if( d > 575 * kFrameWidth)
        {
            d = 575 * kFrameWidth;
        }
        
        memset(m_screen + d, 0, kFrameWidth);
    }
    
    unsigned char *GetLinePtr(int y)
    {
        int d = y * kFrameWidth + ScreenAdjust;
        if( d > 575 * kFrameWidth)
        {
            d = 575 * kFrameWidth;
        }
        
        return((unsigned char *)(m_screen + d));
    }


    void Dump(char *fname)
    {
        FILE *fp = fopen(fname, "wb");
        fwrite(m_screen, 1, kFrameWidth * kFrameHeight, fp);
        fclose(fp);
        
    }
    
    unsigned char *Data()
    {
        return m_screen;
    }
    
private:
    
    unsigned char *m_screen; /* ARGB -- with the alpha ignored */
};


#endif /* defined(__BeebEmX__BeebFrame__) */
