//
//  SheliaMemory.cpp
//  BeebEmX
//
//  Created by Steven Bagley on 16/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#include "6502core.h"
#include "SheilaMemory.h"

#include "sysvia.h"
#include "uservia.h"
#include "serial.h"
#include "video.h"

CSheilaMemory::CSheilaMemory(CBeebMemory *memory) : m_memory(memory)
{
    memset(m_sheilaLUT, 0, sizeof(IMemory*) * 256);
}

void CSheilaMemory::RegisterDevice(int address, IMemory *device)
{
    m_sheilaLUT[address &0xff] = device;
}




int CSheilaMemory::ReadByte(const int address)
{
    int Value = 0xff;
    
    /* VIAs first - games seem to do really heavy reading of these */
    /* Can read from a via using either of the two 16 bytes blocks */
    if ((address & ~0xf)==0xfe40 || (address & ~0xf)==0xfe50)
    {
        SyncIO();
        Value = SysVIARead(address & 0xf);
        AdjustForIORead();
        return Value;
    }
    
    if ((address & ~0xf)==0xfe60 || (address & ~0xf)==0xfe70)
    {
        SyncIO();
        Value = UserVIARead(address & 0xf);
        AdjustForIORead();
        return Value;
    }
    
    if ((address & ~7)==0xfe00) {
        SyncIO();
        Value = CRTCRead(address & 0x7);
        AdjustForIORead();
        return Value;
    }
    
    if (address==0xfe08) {
        SyncIO();
        Value = Read_ACIA_Status();
        AdjustForIORead();
        return Value;
    }
    
    if (address==0xfe09) {
        SyncIO();
        Value = Read_ACIA_Rx_Data();
        AdjustForIORead();
        return Value;
    }
    
    if (address==0xfe10)
    {
        SyncIO();
        Value = Read_SERPROC();
        AdjustForIORead();
        return Value;
    }
    
    if(m_sheilaLUT[address & 0xff])
    {
        return m_sheilaLUT[address &0xff]->ReadByte(address);
    }
    
    
#if 0
    /* Rob: BBC AUG says FE20 is econet stn no. for bbc b. [It's in cmos for a master,]
     This is wrong and doesn't work. Study of the circuit diagram for a model B (IC26) shows it's at
     FE18-FE1F. This now works.  AUG says read station also enables Econet NMI but again the circuit
     shows the read-station select line is also INTOFF. (and it's any access to FE18, not just a read.)
     */
    if (EconetEnabled &&
        ((MachineType!=3 && (address & ~3)==0xfe18) ||
         (MachineType==3 && (address & ~3)==0xfe38)) ) {
            if (DebugEnabled)
                DebugDisplayTrace(DEBUG_ECONET, true, "Econet: INTOFF");
            EconetNMIenabled = INTOFF;
            return(Read_Econet_Station());
        }
#endif
#if 0
    if (address>=0xfe18 && address<=0xfe20 && MachineType==3)
    {
        return(AtoDRead(address - 0xfe18));
    }
#endif
    
    
#if 0
    /* Rob: BBC AUG states video ULA at FE20-21 is write-only - why support reading?
     The circuit diagram shows read of FE20-FE27 is INTON (maybe this is where AUG confusion came from)
     INTON & INTOFF enable/disable NMI from the 68B54 via flip flop IC97, IC27 & IC91.  Guess what happens
     if the ADLC is holding an NMI which was masked by INTOFF and you turn INTON ....!
     (NMI from the FDC is always enabled)
     */
    if (EconetEnabled &&
        ((MachineType!=3 && (address & ~3)==0xfe20) ||
         (MachineType==3 && (address & ~3)==0xfe3c)) ) {
            if (DebugEnabled) DebugDisplayTrace(DEBUG_ECONET, true, "Econet: INTON");
            if (!EconetNMIenabled) {  // was off
                EconetNMIenabled = INTON;  // turn on
                if (ADLC.status1 & 128) {			// irq pending?
                    NMIStatus|=1<<nmi_econet;
                    if (DebugEnabled) DebugDisplayTrace(DEBUG_ECONET, true, "Econet: delayed NMI asserted");
                }
            }
        }
#endif
    
    if ((address & ~3)==0xfe20) {
        return(VideoULARead(address & 0xf));
    }
    
    if ((address & ~3)==0xfe30)
    {
        return m_memory->GetSWR(); /* report back ROMSEL - I'm sure the beeb allows ROMSEL read.. correct me if im wrong. - Richard Gellman */
    }
    
#if 0
    // In the Master at least, ROMSEL/ACCCON seem to be duplicated over a 4 byte block.
    if ((address & ~3)==0xfe34 && machine->MachineType() == 3)
    {
        return(ACCCON);
    }
    if (((address & ~0x1f)==0xfe80) && (MachineType!=3) && (NativeFDC)) {
        return(Disc8271_read(address & 0x7));
    }
#endif
#if 0
    if ((address & ~0x1f)==0xfea0) {
        if (EconetEnabled)
            return(ReadEconetRegister(address & 3)); /* Read 68B54 ADLC */
        return(0xfe); // if not enabled
    }
#endif
#if 0
    if ((address & ~0x1f)==0xfec0 && machine->MachineType()!=3) {
        SyncIO();
        Value = AtoDRead(address & 0xf);
        AdjustForIORead();
        return Value;
    }
#endif
    if ((address & ~0x1f)==0xfee0)
    {
#if 0
        if (TorchTube)
            return(ReadTorchTubeFromHostSide(address&0x1f)); //Read From Torch Tube
        else
            return(ReadTubeFromHostSide(address & 7)); //Read From Tube
#endif
        return 0xfe;
    }
    
#if 0
#endif
    
    return(0xFF);
}

void CSheilaMemory::WriteByte(const int address, const unsigned char value)
{
    /* Can write to a via using either of the two 16 bytes blocks */
    if ((address & ~0xf)==0xfe40 || (address & ~0xf)==0xfe50)
    {
        SyncIO();
        AdjustForIOWrite();
        SysVIAWrite((address & 0xf), value);
        return;
    }
    
    /* Can write to a via using either of the two 16 bytes blocks */
    if ((address & ~0xf)==0xfe60 || (address & ~0xf)==0xfe70)
    {
        SyncIO();
        AdjustForIOWrite();
        UserVIAWrite((address & 0xf), value);
        return;
    }
    
    if ((address & ~0x7) == 0xfe00)
    {
        SyncIO();
        AdjustForIOWrite();
        CRTCWrite(address & 0x7, value);
        return;
    }
    
    if (address==0xfe08)
    {
        SyncIO();
        AdjustForIOWrite();
        Write_ACIA_Control(value);
        return;
    }
    if (address==0xfe09) {
        SyncIO();
        AdjustForIOWrite();
        Write_ACIA_Tx_Data(value);
        return;
    }
    if (address==0xfe10)
    {
        SyncIO();
        AdjustForIOWrite();
        Write_SERPROC(value);
        return;
    }
    
    
    if(m_sheilaLUT[address & 0xff])
    {
        m_sheilaLUT[address &0xff]->WriteByte(address,value);
    }

    
#if 0
    //Rob: econet NMI mask
    if (EconetEnabled && ((machine->MachineType() != 3 && (address & ~3) == 0xfe18) || (machine->MachineType() == 3 && (address & ~3) == 0xfe38)) )
    {
        if(DebugEnabled)
        {
            DebugDisplayTrace(DEBUG_ECONET, true, "Econet: INTOFF(w)");
        }
        EconetNMIenabled = INTOFF;
    }
#endif
    
#if 0
    if ((address & ~0x7)==0xfe18 && machine->MachineType()==3)
    {
        AtoDWrite((address & 0x7),Value);
        return;
    }
#endif
    
    
    if ((address & ~0x3)==0xfe20)
    {
        VideoULAWrite(address & 0xf, value);
        return;
    }
    

    
    if (address>=0xfe30 && address<0xfe34)
    {
        m_memory->SetSWR(value);
        return;
    }
    
#if 0
    // In the Master at least, ROMSEL/ACCCON seem to be duplicated over a 4 byte block.
    /*cerr << "Write *0x" << hex << address << "=0x" << Value << dec << "\n"; */
    if (address>=0xfe34 && address<0xfe38 && machine->MachineType()==3)
    {
        FiddleACCCON(Value);
        return;
    }
    
    if (((address & ~0x1f)==0xfe80) && (machine->MachineType()!=3) && (NativeFDC)) {
        Disc8271_write((address & 7),Value);
        return;
    }
#endif
    
#if 0
    //Rob: add econet
    if (address>=0xfeA0 && address<0xfebf && (EconetEnabled) ) {
        WriteEconetRegister((address & 3), Value);
        return;
    }
#endif
    
#if 0
    if ((address & ~0x1f)==0xfec0 && machine->MachineType()!=3) {
        SyncIO();
        AdjustForIOWrite();
        AtoDWrite((address & 0xf),Value);
        return;
    }
#endif
    
#if 0
    if ((address&~0xf)==0xfee0)
    {
        if (TorchTube)
            WriteTorchTubeFromHostSide(address&0xf,Value);
        else
            WriteTubeFromHostSide(address&7,Value);
    }
#endif
    
    return;
}