//
//  DirectSoundStreamer.h
//  BeebEmX
//
//  Created by Steven Bagley on 18/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#ifndef __BeebEmX__DirectSoundStreamer__
#define __BeebEmX__DirectSoundStreamer__

#include <stdio.h>

class DirectSoundStreamer : public SoundStreamer
{
public:
    struct Fail : std::exception
    {
        char const *what() const throw()
        {
            return "DirectSoundStreamer::Fail";
        }
    };
    
    struct PrimaryUnsupported : Fail
    {
        char const *what() const throw()
        {
            return "DirectSoundStreamer::PrimaryUnsupported";
        }
    };
    
    DirectSoundStreamer( std::size_t rate, bool primary );
    
    ~DirectSoundStreamer()
    {
        if (m_pDirectSoundBuffer)
            m_pDirectSoundBuffer->Release();
        if (m_pDirectSound)
            m_pDirectSound->Release();
    }
    
    std::size_t BufferSize() const
    {
        return m_stream;
    }
    
    void Play()
    {
        m_pDirectSoundBuffer->Play( 0, 0, DSBPLAY_LOOPING );
    }
    
    void Pause()
    {
        m_pDirectSoundBuffer->Stop();
    }
    
    void Stream( void const *pSamples );
    
private:
    IDirectSound* m_pDirectSound;
    IDirectSoundBuffer* m_pDirectSoundBuffer;
    std::size_t m_begin;
    std::size_t m_physical;
    std::size_t m_rate;
    std::size_t m_stream;
    std::size_t m_virtual;
    
    static std::size_t const m_count = 6;
};

class XAudio2Streamer : public SoundStreamer
{
public:
    struct Fail : std::exception
    {
        char const *what() const throw()
        {
            return "XAudio2Streamer::Fail";
        }
    };
    
    XAudio2Streamer( std::size_t rate );
    
    ~XAudio2Streamer()
    {
        if (m_pXAudio2)
            m_pXAudio2->Release();
    }
    
    std::size_t BufferSize() const
    {
        return m_size;
    }
    
    void Play()
    {
        m_pXAudio2SourceVoice->Start( 0 );
    }
    
    void Pause()
    {
        m_pXAudio2SourceVoice->Stop( 0 );
    }
    
    void Stream( void const *pSamples );
    
private:
    typedef unsigned char Sample;
    typedef std::vector< Sample > Samples;
    Samples m_samples;
    IXAudio2* m_pXAudio2;
    IXAudio2MasteringVoice *m_pXAudio2MasteringVoice;
    IXAudio2SourceVoice *m_pXAudio2SourceVoice;
    std::size_t m_index;
    std::size_t m_rate;
    std::size_t m_size;
    
    static std::size_t const m_count = 4;
};



#endif /* defined(__BeebEmX__DirectSoundStreamer__) */
