/****************************************************************
BeebEm - BBC Micro and Master 128 Emulator
Copyright (C) 1994  David Alan Gilbert
Copyright (C) 1997  Mike Wyatt
Copyright (C) 2001  Richard Gellman
Copyright (C) 2008  Rich Talbot-Watkins

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public 
License along with this program; if not, write to the Free 
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
****************************************************************/

/* Win32 port - Mike Wyatt 7/6/97 */
/* Conveted Win32 port to use DirectSound - Mike Wyatt 11/1/98 */

#include "beebsound.h"

#include <iostream>
#include <fstream>
#include <math.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "6502core.h"
#include "port.h"
#include "beebwin.h"
#ifdef SPEECH_ENABLED
#include "speech.h"
#endif

#include "CoreAudioStreamer.h"

#include <exception>
#include <vector>

#if 0
extern AVIWriter *aviWriter;
#endif
//  #define DEBUGSOUNDTOFILE

#define PREFSAMPLERATE 44100
#define MAXBUFSIZE 32768

static unsigned char SoundBuf[MAXBUFSIZE];
unsigned char SpeechBuf[MAXBUFSIZE];


SoundConfig::Option SoundConfig::Selection;
int UsePrimaryBuffer=0;
int SoundEnabled = 1;
int SoundChipEnabled = 1;
int SoundSampleRate = PREFSAMPLERATE;
int SoundVolume = 3;
int SoundAutoTriggerTime;
int SoundBufferSize;
double CSC[4]={0,0,0,0},CSA[4]={0,0,0,0}; // ChangeSamps Adjusts
char SoundExponentialVolume = 1;

/* Number of places to shift the volume */
#define VOLMAG 3

int Speech[4];

FILE *sndlog=NULL;

volatile struct
{
    unsigned int ToneFreq[4];
    unsigned int ChangeSamps[4]; /* How often this channel should flip its otuput */
    unsigned int ToneVolume[4]; /* In units of /dev/dsp */
    struct
    {
        unsigned int FB:1; /* =0 for periodic, =1 for white */
        unsigned int Freq:2; /* 0=low, 1=medium, 2=high, 3=tone gen 1 freq */
        unsigned int Vol:4;
    } Noise;
    int LastToneFreqSet; /* the tone generator last set - for writing the 2nd byte */
} BeebState76489;


int RealVolumes[4]; // Holds the real volume values for state save use

static int ActiveChannel[4] = {false, false, false, false}; /* Those channels with non-0 voolume */

// Set it to an array for more accurate sound generation

static unsigned int samplerate;
static double OurTime=0.0; /* Time in sample periods */

CycleCountT SoundTrigger; /* Time to trigger a sound event */

static unsigned int GenIndex[4]; /* Used by the voice generators */
static int GenState[4];
static int bufptr=0;
int SoundDefault = 1;
double SoundTuning=0.0; // Tunning offset

void PlayUpTil(double DestTime);
int GetVol(int vol);
bool bReRead = false;
volatile bool bDoSound = true;
int WriteOffset=0;
#if 0
LARGE_INTEGER PFreq,LastPCount,CurrentPCount;
#endif
double CycleRatio;
struct AudioType TapeAudio; // Tape audio decoder stuff
bool TapeSoundEnabled;
int PartSamples=1;
bool Playing=0;

#if 0


class CRawStreamer : public SoundStreamer
{
public:
    
    CRawStreamer()
    {
        m_fp = fopen("/Users/steve/Desktop/sound.raw", "wb");
    }
    
    virtual ~CRawStreamer()
    {
        fclose(m_fp);
    }
    
    virtual size_t BufferSize() const
    {
        return 44100 * 20 / 1000;
    }
    
    virtual void Play()
    {
    
        
    }
    virtual void Pause()
    {
        
        
    }
    
    virtual void Stream( void const *pSamples )
    {
        fwrite(pSamples, 1, 44100 * 20 / 1000, m_fp);
        
    }
    
private:
    FILE *m_fp;
    
};
#endif


SoundStreamer *pSoundStreamer = NULL;

/****************************************************************************/
/* Writes sound data to a sound buffer */
int WriteToSoundBuffer(unsigned char *lpbSoundData)
{
	if( pSoundStreamer )
		pSoundStreamer->Stream( lpbSoundData );
    
#if 0
	if( aviWriter )
	{
		HRESULT hResult( aviWriter->WriteSound( lpbSoundData, SoundBufferSize ) );
		if( FAILED( hResult ) && hResult != E_UNEXPECTED )
		{
			MessageBox( GETHWND, "Failed to write sound to AVI file", "BeebEm", MB_OK | MB_ICONERROR );
			delete aviWriter;
			aviWriter = 0;
		}
	}
#endif
	return true;
}


/****************************************************************************/
/* DestTime is in samples */
void PlayUpTil(double DestTime)
{
	int tmptotal,channel,bufinc,tapetotal;
	char Extras;
	int SpeechPtr = 0;
	int i;

#ifdef SPEECH_ENABLED
	if (MachineType != 3 && SpeechEnabled)
	{
		SpeechPtr = 0;
		memset(SpeechBuf, 128, sizeof(SpeechBuf));
		int len = (int) (DestTime - OurTime + 1);
		if (len > MAXBUFSIZE)
			len = MAXBUFSIZE;
		tms5220_update(SpeechBuf, len);
	}
#endif

	while (DestTime > OurTime)
    {

		for(bufinc = 0; (bufptr < SoundBufferSize) && ((OurTime + bufinc) < DestTime); bufptr++, bufinc++)
        {
			int tt;
			tmptotal=0;
			Extras=4;

			if (SoundChipEnabled)
            {
				// Begin of for loop
				for(channel=1;channel<=3;channel++)
                {
					if (ActiveChannel[channel])
                    {
						if ((GenState[channel]) && (!Speech[channel]))
                        {
							tmptotal+=BeebState76489.ToneVolume[channel];
                        }
                        
                        if ((!GenState[channel]) && (!Speech[channel]))
                        {
							tmptotal-=BeebState76489.ToneVolume[channel];
                        }
                        
                        if(Speech[channel])
                        {
							tmptotal+=(BeebState76489.ToneVolume[channel]-GetVol(7));
                        }
                        
                        GenIndex[channel]++;
						
                        tt = (int)CSC[channel];
						
                        if (!PartSamples)
                        {
                            tt=0;
                        }
                        
						if(GenIndex[channel] >= (BeebState76489.ChangeSamps[channel]+tt))
                        {
							if (CSC[channel] >= 1.0)
                            {
								CSC[channel]-=1.0;
                            }
                            
                            CSC[channel]+=CSA[channel];
							
                            GenIndex[channel]=0;
							GenState[channel]^=1;
						}
					}
				} /* Channel loop */
                
                /* Now put in noise generator stuff */
                if (ActiveChannel[0]) {
                    if (BeebState76489.Noise.FB) {
                        /* White noise */
                        if (GenState[0])
                            tmptotal+=BeebState76489.ToneVolume[0];
                        else
                            tmptotal-=BeebState76489.ToneVolume[0];
                        GenIndex[0]++;
                        switch (BeebState76489.Noise.Freq)
                        {
                            case 0: /* Low */
                                if (GenIndex[0]>=(samplerate/10000))
                                {
                                    GenIndex[0]=0;
                                    GenState[0]=rand() & 1;
                                }
                                break;
                                
                            case 1: /* Med */
                                if (GenIndex[0]>=(samplerate/5000))
                                {
                                    GenIndex[0]=0;
                                    GenState[0]=rand() & 1;
                                }
                                break;
                                
                            case 2: /* High */
                                if (GenIndex[0]>=(samplerate/2500))
                                {
                                    GenIndex[0]=0;
                                    GenState[0]=rand() & 1;
                                }
                                break;
                                
                            case 3: /* as channel 1 */
                                if (GenIndex[0]>=BeebState76489.ChangeSamps[1])
                                {
                                    GenIndex[0]=0;
                                    GenState[0]=rand() & 1;
                                }
                                break;
                        } /* Freq type switch */
                    }
                    else
                    {
                        /* Periodic */
                        if (GenState[0])
                            tmptotal+=BeebState76489.ToneVolume[0];
                        else
                            tmptotal-=BeebState76489.ToneVolume[0];
                        
                        GenIndex[0]++;
                        
                        switch (BeebState76489.Noise.Freq)
                        {
                            case 2: /* Low */
                                if (GenState[0])
                                {
                                    if (GenIndex[0]>=(samplerate/125))
                                    {
                                        GenIndex[0]=0;
                                        GenState[0]=0;
                                    }
                                }
                                else
                                {
                                    if (GenIndex[0]>=(samplerate / 1250))
                                    {
                                        GenIndex[0]=0;
                                        GenState[0]=1;
                                    }
                                }
                                break;
                                
                            case 1: /* Med */
                                if (GenState[0])
                                {
                                    if (GenIndex[0]>=(samplerate / 250))
                                    {
                                        GenIndex[0]=0;
                                        GenState[0]=0;
                                    }
                                }
                                else
                                {
                                    if (GenIndex[0]>=(samplerate / 2500))
                                    {
                                        GenIndex[0]=0;
                                        GenState[0]=1;
                                    }
                                }
                                break;
                                
                            case 0: /* High */
                                if (GenState[0])
                                {
                                    if (GenIndex[0]>=(samplerate/500))
                                    {
                                        GenIndex[0]=0;
                                        GenState[0]=0;
                                    }
                                }
                                else
                                {
                                    if (GenIndex[0]>=(samplerate/5000))
                                    {
                                        GenIndex[0]=0;
                                        GenState[0]=1;
                                    }
                                }
                                break;
                                
                            case 3: /* Tone gen 1 */
                                tt=(int)CSC[0];
                                if (GenIndex[0] >= (BeebState76489.ChangeSamps[1]+tt))
                                {
                                    static int per=0;
                                    CSC[0]+=CSA[1]-tt;
                                    GenIndex[0] = 0;
                                    GenState[0] = (per == 0);
                                    if (++per==30)
                                    {
                                        per=0;
                                    }
                                }
                                break; 
                                
                        } /* Freq type switch */
                    }
                }
            }

#ifdef SPEECH_ENABLED
			// Mix in speech sound
			if (SpeechEnabled) if (MachineType != 3) tmptotal += (SpeechBuf[SpeechPtr++]-128)*10;
#endif

#if 0
			// Mix in sound samples here
			for (i = 0; i < NUM_SOUND_SAMPLES; ++i)
            {
				if (SoundSamples[i].playing)
                {
					tmptotal+=(SoundSamples[i].pBuf[SoundSamples[i].pos]-128)*10;
					Extras++;
					SoundSamples[i].pos += (44100 / samplerate);
					if (SoundSamples[i].pos >= SoundSamples[i].len) {
						if (SoundSamples[i].repeat)
							SoundSamples[i].pos = 0;
						else
							SoundSamples[i].playing = false;
					}
				}
			}
#endif
			if (TapeSoundEnabled)
            {
				// Mix in tape sound here
				tapetotal=0; 
				if ((TapeAudio.Enabled) && (TapeAudio.Signal==2))
                {
					if (TapeAudio.Samples++ >= 36)
                    {
                        TapeAudio.Samples=0;
                    }
                    
                    tapetotal = (int)(sin(((TapeAudio.Samples * 20) * 3.14) / 180) * 80);
					Extras++;
				}
                
				if ((TapeAudio.Enabled) && (TapeAudio.Signal==1))
                {
					tapetotal=(int)(sin(((TapeAudio.Samples*(10*(1+TapeAudio.CurrentBit)))*3.14)/180)*(80+(40*(1-TapeAudio.CurrentBit))));
					// And if you can follow that equation, "ill give you the money meself" - Richard Gellman
					if (TapeAudio.Samples++>=36) {
						TapeAudio.Samples=0;
						TapeAudio.BytePos++;
						if (TapeAudio.BytePos<=10) TapeAudio.CurrentBit=(TapeAudio.Data & (1<<(10-TapeAudio.BytePos)))?1:0;
					}
					if (TapeAudio.BytePos>10) {
						TapeAudio.ByteCount--;
						if (!TapeAudio.ByteCount) TapeAudio.Signal=2; else { TapeAudio.BytePos=1; TapeAudio.CurrentBit=0; }
					}
					Extras++;
				}
				tmptotal+=tapetotal;
			}

			/* Make it a bit louder under Windows */
			if (Extras)
				tmptotal/=Extras;
			else
				tmptotal=0;

            SoundBuf[bufptr] = (tmptotal / SoundVolume) ;// + 128;
		} /* buffer loop */

		/* Only write data when buffer is full */
		if (bufptr == SoundBufferSize)
		{
#ifdef DEBUGSOUNDTOFILE
			FILE *fd = fopen("/audio.dbg", "a+b");
			if (fd != NULL)
			{
				fwrite(SoundBuf, 1, SoundBufferSize, fd);
				fclose(fd);
			}
			else
			{
				MessageBox(GETHWND,"Failed to open audio.dbg",WindowTitle,MB_OK|MB_ICONERROR);
				exit(1);
			}
#else
			//if (sndlog!=NULL)
				//fprintf(sndlog,"Sound write at %lf Samples\n",DestTime);
				//fwrite(SoundBuf,1,SoundBufferSize,sndlog);
			int hr;
			hr = WriteToSoundBuffer(SoundBuf);
#endif
			// buffer swapping no longer needed
			bufptr=0;
		}

		OurTime += bufinc;
	} /* While time */ 
}
/****************************************************************************/
/* Convert time in cycles to time in samples                                */

static CycleCountT LastBeebCycle = 0; /* Last parameter to this function */
static double LastOurTime = 0; /* Last result of this function */

static double CyclesToSamples(CycleCountT BeebCycles)
{
    double tmp;
    
    /* OK - beeb cycles are in 2MHz units, ours are in 1/samplerate */
    /* This is all done incrementally - find the number of ticks since the last call in both domains.  This does mean this should only be called once */
    /* Extract number of cycles since last call */

    if (BeebCycles < LastBeebCycle)
    {
        /* Wrap around in beebs time */
        tmp=((double)CycleCountWrap-(double)LastBeebCycle)+(double)BeebCycles;
    }
    else
    {
        tmp=(double)BeebCycles-(double)LastBeebCycle;
    }
    
#if 0
    tmp /= (mainWin->m_RealTimeTarget) ? mainWin->m_RealTimeTarget : 1;
#endif
    
    /*fprintf(stderr,"Convert tmp=%f\n",tmp); */
    LastBeebCycle = BeebCycles;
    
    tmp *= (samplerate);
    tmp /= 2000000.0; /* Few - glad thats a double! */
    
    LastOurTime += tmp;
    return LastOurTime;
}

/****************************************************************************/
static void InitAudioDev(int sampleratein)
{

	delete pSoundStreamer;
	samplerate = sampleratein;
    
    
    pSoundStreamer = new CCoreAudioStreamer();
    SoundEnabled = 1;
    
    
#if 0
    if( SoundConfig::Selection == SoundConfig::XAudio2 )
	{
		try
		{
			pSoundStreamer = new XAudio2Streamer( samplerate );
			return;
		}
		catch( ... )
		{
		}
		SoundConfig::Selection = SoundConfig::DirectSound;
	}
	if( UsePrimaryBuffer )
	{
		try
		{
			pSoundStreamer = new DirectSoundStreamer( samplerate, true );
			return;
		}
		catch( DirectSoundStreamer::PrimaryUnsupported const & )
		{
			MessageBox(GETHWND, "Use of primary DirectSound buffer unsupported on this system. Using secondary DirectSound buffer instead", WindowTitle, MB_OK | MB_ICONERROR);
			UsePrimaryBuffer = 0;
			mainWin->SetPBuff();
		}
		catch( ... )
		{
		}
	}
	if( ! UsePrimaryBuffer )
	{
		try
		{
			pSoundStreamer = new DirectSoundStreamer( samplerate, false );
			return;
		}
		catch( ... )
		{
		}
	}
	SoundEnabled = 0;
	MessageBox( GETHWND, "Attempt to start sound system failed", "BeebEm", MB_OK | MB_ICONERROR );
#endif
}// InitAudioDev



/****************************************************************************/
/* The 'freqval' variable is the value as sene by the 76489                 */
static void SetFreq(int Channel, int freqval) {
  unsigned int freq;
  int ChangeSamps; /* Number of samples after which to change */
  //fprintf(sndlog,"Channel %d - Value %d\n",Channel,freqval);
  double t;

  if (freqval==0) freqval=1;
  if (freqval<5) Speech[Channel]=1; else Speech[Channel]=0;
  freq=4000000/(32*freqval);

  t=( (( (double)samplerate/(double)freq)/2.0) +SoundTuning);
  ChangeSamps=(int)t;
  CSA[Channel]=(double)(t-ChangeSamps);
  CSC[Channel]=CSA[Channel];  // we look ahead, so should already include the fraction on the first update
  if (Channel==1) {
    CSC[0]=CSC[1];
  }

  BeebState76489.ChangeSamps[Channel]=ChangeSamps;
};

/****************************************************************************/
static void SoundTrigger_Real(void)
{
  double nowsamps;
  nowsamps = CyclesToSamples(TotalCycles);
  PlayUpTil(nowsamps);
  SoundTrigger = TotalCycles + SoundAutoTriggerTime;
}; /* SoundTrigger_Real */

void Sound_Trigger(int NCycles)
{
    if (SoundTrigger<=TotalCycles)
    {
        SoundTrigger_Real();
    }
    // fprintf(sndlog,"SoundTrigger_Real was called from Sound_Trigger\n"); }
}

void SoundChipReset(void)
{
    BeebState76489.LastToneFreqSet = 0;
    BeebState76489.ToneVolume[0] = 0;
    BeebState76489.ToneVolume[1] = BeebState76489.ToneVolume[2] = BeebState76489.ToneVolume[3] = GetVol(15);
    BeebState76489.ToneFreq[0] = BeebState76489.ToneFreq[1] = BeebState76489.ToneFreq[2] = 1000;
    BeebState76489.ToneFreq[3] = 1000;
    BeebState76489.Noise.FB = 0;
    BeebState76489.Noise.Freq = 0;
    ActiveChannel[0] = false;
    ActiveChannel[1] = ActiveChannel[2] = ActiveChannel[3] = false;
}

/****************************************************************************/
/* Called to enable sound output                                            */
void SoundInit()
{
    ClearTrigger(SoundTrigger);
  
    LastBeebCycle = TotalCycles;
    LastOurTime = (double)LastBeebCycle * (double)SoundSampleRate / 2000000.0;
    OurTime = LastOurTime;
    bufptr = 0;

    InitAudioDev(SoundSampleRate);
    
    if (SoundSampleRate == 44100)
    {
        SoundAutoTriggerTime = 5000;
    }
    
    if (SoundSampleRate == 22050)
    {
        SoundAutoTriggerTime = 10000;
    }
    
    if (SoundSampleRate == 11025)
    {
        SoundAutoTriggerTime = 20000;
    }
    
    SoundBufferSize = pSoundStreamer ? pSoundStreamer->BufferSize() : SoundSampleRate / 50;
    
#if 0
    LoadSoundSamples();
#endif
    bReRead = true;
    SoundTrigger=TotalCycles+SoundAutoTriggerTime;
}

void SwitchOnSound(void)
{
  SetFreq(3,1000);
  ActiveChannel[3]=true;
  BeebState76489.ToneVolume[3]=GetVol(15);
}

void SetSound(char State)
{
	if( pSoundStreamer )
	{
		switch( State )
		{
		case MUTED:
			pSoundStreamer->Pause();
			break;
		case UNMUTED:
			pSoundStreamer->Play();
			break;
		}
	}
}


/****************************************************************************/
/* Called to disable sound output                                           */
void SoundReset(void)
{

	delete pSoundStreamer;
	pSoundStreamer = 0;

	ClearTrigger(SoundTrigger);

}//	SoundReset

/****************************************************************************/
/* Called in sysvia.cc when a write is made to the 76489 sound chip         */
void Sound_RegWrite(int value)
{
    int trigger = 0;
    unsigned char VolChange;
    
    if (!SoundEnabled)
    {
        return;
    }
    
    VolChange = 4;
    
    if (!(value & 0x80))
    {
        unsigned val = BeebState76489.ToneFreq[BeebState76489.LastToneFreqSet] & 15;
        
        /* Its changing the top half of the frequency */
        val |= (value & 0x3f) << 4;
        
        /* And update */
        BeebState76489.ToneFreq[BeebState76489.LastToneFreqSet] = val;
        SetFreq(BeebState76489.LastToneFreqSet+1,BeebState76489.ToneFreq[BeebState76489.LastToneFreqSet]);
        trigger = 1;
    }
    else
    {
        /* Another register */
        VolChange = 0xff;
        switch ((value>>4) & 0x7)
        {
            case 0: /* Tone 3 freq */
                BeebState76489.ToneFreq[2] = (BeebState76489.ToneFreq[2] & 0x3f0) | (value & 0xf);
                SetFreq(3,BeebState76489.ToneFreq[2]);
                BeebState76489.LastToneFreqSet = 2;
                //	trigger = 1;
                break;
                
            case 1: /* Tone 3 vol */
                RealVolumes[3] = value & 15;
                if ((BeebState76489.ToneVolume[3]==0) && ((value &15)!=15)) ActiveChannel[3]=true;
                if ((BeebState76489.ToneVolume[3]!=0) && ((value &15)==15)) ActiveChannel[3]=false;
                BeebState76489.ToneVolume[3]=GetVol(15-(value & 15));
                BeebState76489.LastToneFreqSet=2;
                trigger = 1;
                VolChange=3;
                break;
                
            case 2: /* Tone 2 freq */
                BeebState76489.ToneFreq[1]=(BeebState76489.ToneFreq[1] & 0x3f0) | (value & 0xf);
                BeebState76489.LastToneFreqSet=1;
                SetFreq(2,BeebState76489.ToneFreq[1]);
                //	trigger = 1;
                break;
                
            case 3: /* Tone 2 vol */
                RealVolumes[2]=value&15;
                if ((BeebState76489.ToneVolume[2]==0) && ((value &15)!=15)) ActiveChannel[2]=true;
                if ((BeebState76489.ToneVolume[2]!=0) && ((value &15)==15)) ActiveChannel[2]=false;
                BeebState76489.ToneVolume[2]=GetVol(15-(value & 15));
                BeebState76489.LastToneFreqSet=1;
                trigger = 1;
                VolChange=2;
                break;
                
            case 4: /* Tone 1 freq (Possibly also noise!) */
                BeebState76489.ToneFreq[0]=(BeebState76489.ToneFreq[0] & 0x3f0) | (value & 0xf);
                BeebState76489.LastToneFreqSet=0;
                SetFreq(1,BeebState76489.ToneFreq[0]);
                //	trigger = 1;
                break;
                
            case 5: /* Tone 1 vol */
                RealVolumes[1]=value&15;
                if ((BeebState76489.ToneVolume[1] == 0) && ((value & 15) != 15))
                {
                    ActiveChannel[1] = true;
                }
                if ((BeebState76489.ToneVolume[1] != 0) && ((value & 15) == 15))
                {
                    ActiveChannel[1] = false;
                }
                BeebState76489.ToneVolume[1]=GetVol(15-(value & 15));
                BeebState76489.LastToneFreqSet=0;
                trigger = 1;
                VolChange=1;
                break;
                
            case 6: /* Noise control */
                BeebState76489.Noise.Freq=value &3;
                BeebState76489.Noise.FB=(value>>2)&1;
                
                trigger = 1;
                break;
                
            case 7: /* Noise volume */
                if ((BeebState76489.ToneVolume[0]==0) && ((value &15)!=15)) ActiveChannel[0]=true;
                if ((BeebState76489.ToneVolume[0]!=0) && ((value &15)==15)) ActiveChannel[0]=false;
                RealVolumes[0]=value&15;
                BeebState76489.ToneVolume[0]=GetVol(15-(value & 15));
                trigger = 1;
                VolChange=0;
                break;
        };
        //if (VolChange<4) fprintf(sndlog,"Channel %d - Volume %d at %lu Cycles\n",VolChange,value &15,SoundCycles);
    }
    if (trigger)
    {
        SoundTrigger_Real();
    }
}



void DumpSound(void) {
	
}

int GetVol(int vol) {
	if (SoundExponentialVolume)	{
//		static int expVol[] = { 0,  2,  4,  6,  9, 12, 15, 19, 24, 30, 38, 48, 60, 76,  95, 120 };
		static int expVol[] = { 0, 11, 14, 17, 20, 24, 28, 33, 39, 46, 54, 63, 74, 87, 102, 120 };
		if (vol >= 0 && vol <= 15)
			return expVol[vol];
		else
			return 0;
	}
	else {
		return vol << VOLMAG;
	}
}

#if 0

void LoadSoundUEF(FILE *SUEF) {
	// Sound block
	unsigned char Chan;
	int Data;
	int RegVal; // This will be filled in by the data processor
	for (Chan=1;Chan<4;Chan++) {
		Data=fget16(SUEF);
		// Send the data direct to Sound_RegWrite()
		RegVal=(((Chan-1)*2)<<4)|128;
		RegVal|=(Data&15);
		Sound_RegWrite(RegVal);
		RegVal=(Data&1008)>>4;
		Sound_RegWrite(RegVal);
	}
	for (Chan=1;Chan<4;Chan++) {
		Data=fgetc(SUEF);
		RegVal=((((Chan-1)*2)+1)<<4)|128;
		RegVal|=Data&15;
		Sound_RegWrite(RegVal);
		BeebState76489.ToneVolume[4-Chan]=GetVol(15-Data);
		if (Data!=15) ActiveChannel[4-Chan]=TRUE; else ActiveChannel[4-Chan]=FALSE;
	}
	RegVal=224|(fgetc(SUEF)&7);
	Sound_RegWrite(RegVal);
	Data=fgetc(SUEF);
	RegVal=240|(Data&15);
	Sound_RegWrite(RegVal);
	BeebState76489.ToneVolume[0]=GetVol(15-Data);
	if (Data!=15) ActiveChannel[0]=TRUE; else ActiveChannel[0]=FALSE;
	BeebState76489.LastToneFreqSet=fgetc(SUEF);
	GenIndex[0]=fget16(SUEF);
	GenIndex[1]=fget16(SUEF);
	GenIndex[2]=fget16(SUEF);
	GenIndex[3]=fget16(SUEF);
}

void SaveSoundUEF(FILE *SUEF) {
	unsigned char Noise;
	fput16(0x046B,SUEF);
	fput32(20,SUEF);
	// Sound Block
	fput16(BeebState76489.ToneFreq[2],SUEF);
	fput16(BeebState76489.ToneFreq[1],SUEF);
	fput16(BeebState76489.ToneFreq[0],SUEF);
	fputc(RealVolumes[3],SUEF);
	fputc(RealVolumes[2],SUEF);
	fputc(RealVolumes[1],SUEF);
    Noise=BeebState76489.Noise.Freq |
          (BeebState76489.Noise.FB<<2);
	fputc(Noise,SUEF);
	fputc(RealVolumes[0],SUEF);
	fputc(BeebState76489.LastToneFreqSet,SUEF);
	fput16(GenIndex[0],SUEF);
	fput16(GenIndex[1],SUEF);
	fput16(GenIndex[2],SUEF);
	fput16(GenIndex[3],SUEF);
}

#endif