/****************************************************************
 BeebEm - BBC Micro and Master 128 Emulator
 Copyright (C) 1994  David Alan Gilbert
 Copyright (C) 1994  Nigel Magnay
 Copyright (C) 1997  Mike Wyatt
 Copyright (C) 2001  Richard Gellman
 Copyright (C) 2008  Rich Talbot-Watkins
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 Boston, MA  02110-1301, USA.
 ****************************************************************/

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "BeebFrame.h"
#include "FrameQueue.h"
#include "VideoState.h"

//#define BEEB_DOTIME

#include "6502core.h"
#include "BeebMemory.h"
/*
 #include "beebwin.h"
 */
#if 0
#include "main.h"
#endif
#include "sysvia.h"
#include "video.h"
#if 0
#include "uefstate.h"
#include "beebsound.h"
#endif
#include "debug.h"
#if 0
#include "teletext.h"
#endif
#ifdef BEEB_DOTIME
#include <sys/times.h>
#ifdef SUNOS
#include <sys/param.h>
#endif
#ifdef HPUX
#include <unistd.h>
#endif
#endif

extern CBeebFrame *currentFrame;


typedef union EightUChars
{
    unsigned int data[8];
    ///    EightByteType eightbyte;
} EightUChars;

typedef union SixteenUChars
{
    unsigned int data[16];
    //    EightByteType eightbytes[2];
} SixteenUChars;

extern VideoStateT VideoState;

/* Bit assignments in control reg:
 0 - Flash colour (0=first colour, 1=second)
 1 - Teletext select (0=on chip serialiser, 1=teletext)
 2,3 - Bytes per line (2,3=1,1 is 80, 1,0=40, 0,1=20, 0,0=10)
 4 - CRTC Clock chip select (0 = low frequency, 1= high frequency)
 5,6 - Cursor width in bytes (0,0 = 1 byte, 0,1=not defined, 1,0=2, 1,1=4)
 7 - Master cursor width (if set causes large cursor)
 */

EightUChars FastTable[256];
SixteenUChars FastTableDWidth[256]; /* For mode 4,5,6 */
int FastTable_Valid=0;

typedef void (*LineRoutinePtr)(void);
LineRoutinePtr LineRoutine;

/* Translates middle bits of VideoULA_ControlReg to number of colours */
static int NColsLookup[] = { 16, 4, 2, 0 /* Not supported 16? */, 0, 16, 4, 2}; /* Based on AUG 379 */

unsigned int EM7Font[3][96][20]; // 20 rows to account for "half pixels"

static void LowLevelDoScanLineNarrow();
static void LowLevelDoScanLineWide();
static void LowLevelDoScanLineNarrowNot4Bytes();
static void LowLevelDoScanLineWideNot4Bytes();

/*-------------------------------------------------------------------------------------------------------------*/
void BuildMode7Font(const unsigned char *fontData)
{
    unsigned char m7cc, m7cy;
    unsigned int m7cb;
    unsigned int row1, row2, row3; // row builders for mode 7 graphics
    const unsigned char *p = fontData;
    
    for(m7cc = 32; m7cc <= 127; m7cc++)
    {
        for (m7cy = 0; m7cy <= 17; m7cy++)
        {
            m7cb =  (*p++);
            m7cb |= (*p++) << 8;
            EM7Font[0][m7cc - 32][m7cy + 2] = m7cb << 2;
            EM7Font[1][m7cc - 32][m7cy + 2] = m7cb << 2;
            EM7Font[2][m7cc - 32][m7cy + 2] = m7cb << 2;
        }
        EM7Font[0][m7cc-32][0]=EM7Font[1][m7cc-32][0]=EM7Font[2][m7cc-32][0]=0;
        EM7Font[0][m7cc-32][1]=EM7Font[1][m7cc-32][1]=EM7Font[2][m7cc-32][1]=0;
    }
    
    // Now fill in the graphics - this is built from an algorithm, but has certain lines/columns
    // blanked for separated graphics.
    for (m7cc = 0; m7cc < 96; m7cc++)
    {
        // here's how it works: top two blocks: 1 & 2
        // middle two blocks: 4 & 8
        // bottom two blocks: 16 & 64
        // its only a grpahics character if bit 5 (32) is clear.
        if (!(m7cc & 32))
        {
            row1 = 0;
            row2 = 0;
            row3 = 0;
            
            // left block has a value of 4032, right 63 and both 4095
            if (m7cc & 1) row1|=4032;
            if (m7cc & 2) row1|=63;
            if (m7cc & 4) row2|=4032;
            if (m7cc & 8) row2|=63;
            if (m7cc & 16) row3|=4032;
            if (m7cc & 64) row3|=63;
            // now input these values into the array
            // top row of blocks - continuous
            EM7Font[1][m7cc][0]=EM7Font[1][m7cc][1]=EM7Font[1][m7cc][2]=row1;
            EM7Font[1][m7cc][3]=EM7Font[1][m7cc][4]=EM7Font[1][m7cc][5]=row1;
            // Separated
            row1&=975; // insert gaps
            EM7Font[2][m7cc][0]=EM7Font[2][m7cc][1]=EM7Font[2][m7cc][2]=row1;
            EM7Font[2][m7cc][3]=row1; EM7Font[2][m7cc][4]=EM7Font[2][m7cc][5]=0;
            // middle row of blocks - continuous
            EM7Font[1][m7cc][6]=EM7Font[1][m7cc][7]=EM7Font[1][m7cc][8]=row2;
            EM7Font[1][m7cc][9]=EM7Font[1][m7cc][10]=EM7Font[1][m7cc][11]=row2;
            EM7Font[1][m7cc][12]=EM7Font[1][m7cc][13]=row2;
            // Separated
            row2&=975; // insert gaps
            EM7Font[2][m7cc][6]=EM7Font[2][m7cc][7]=EM7Font[2][m7cc][8]=row2;
            EM7Font[2][m7cc][9]=EM7Font[2][m7cc][10]=EM7Font[2][m7cc][11]=row2;
            EM7Font[2][m7cc][12]=EM7Font[2][m7cc][13]=0;
            // Bottom row - continuous
            EM7Font[1][m7cc][14]=EM7Font[1][m7cc][15]=EM7Font[1][m7cc][16]=row3;
            EM7Font[1][m7cc][17]=EM7Font[1][m7cc][18]=EM7Font[1][m7cc][19]=row3;
            // Separated
            row3&=975; // insert gaps
            EM7Font[2][m7cc][14]=EM7Font[2][m7cc][15]=EM7Font[2][m7cc][16]=row3;
            EM7Font[2][m7cc][17]=row3; EM7Font[2][m7cc][18]=EM7Font[2][m7cc][19]=0;
        } // check for valid char to modify
    } // character loop.
} /* BuildMode7Font */


/*-------------------------------------------------------------------------------------------------------------*/
static void DoFastTable16(void)
{
    unsigned int beebpixvl,beebpixvr;
    unsigned int bplvopen,bplvtotal;
    unsigned char tmp;
    
    for(beebpixvl = 0; beebpixvl < 16; beebpixvl++)
    {
        bplvopen = ((beebpixvl & 8) ? 128 : 0) | ((beebpixvl & 4) ? 32 : 0) | ((beebpixvl & 2) ? 8 : 0) | ((beebpixvl & 1) ? 2 : 0);
        for(beebpixvr = 0; beebpixvr < 16; beebpixvr++)
        {
            bplvtotal = bplvopen | ((beebpixvr & 8) ? 64 : 0) | ((beebpixvr & 4) ? 16 : 0) | ((beebpixvr & 2) ? 4 : 0) | ((beebpixvr & 1) ? 1 : 0);
            tmp = VideoULA_Palette[beebpixvl];
            if(tmp > 7)
            {
                tmp &= 7;
                if (VideoULA_ControlReg & 1)
                {
                    tmp ^= 7;
                }
            }
            
            FastTable[bplvtotal].data[0] = FastTable[bplvtotal].data[1] =
            FastTable[bplvtotal].data[2] = FastTable[bplvtotal].data[3] = tmp;
            
            tmp=VideoULA_Palette[beebpixvr];
            if (tmp>7)
            {
                tmp &= 7;
                if (VideoULA_ControlReg & 1)
                {
                    tmp^=7;
                }
            }
            
            FastTable[bplvtotal].data[4] = FastTable[bplvtotal].data[5] =
            FastTable[bplvtotal].data[6] = FastTable[bplvtotal].data[7] = tmp;
        }; /* beebpixr */
    } /* beebpixl */
} /* DoFastTable16 */

/*-------------------------------------------------------------------------------------------------------------*/
static void DoFastTable16XStep8(void)
{
    unsigned int beebpixvl,beebpixvr;
    unsigned int bplvopen,bplvtotal;
    unsigned char tmp;
    
    for(beebpixvl=0;beebpixvl<16;beebpixvl++) {
        bplvopen=((beebpixvl & 8)?128:0) |
        ((beebpixvl & 4)?32:0) |
        ((beebpixvl & 2)?8:0) |
        ((beebpixvl & 1)?2:0);
        for(beebpixvr=0;beebpixvr<16;beebpixvr++) {
            bplvtotal=bplvopen |
            ((beebpixvr & 8)?64:0) |
            ((beebpixvr & 4)?16:0) |
            ((beebpixvr & 2)?4:0) |
            ((beebpixvr & 1)?1:0);
            tmp=VideoULA_Palette[beebpixvl];
            if (tmp>7) {
                tmp&=7;
                if (VideoULA_ControlReg & 1) tmp^=7;
            };
            FastTableDWidth[bplvtotal].data[0]=FastTableDWidth[bplvtotal].data[1]=
            FastTableDWidth[bplvtotal].data[2]=FastTableDWidth[bplvtotal].data[3]=
            FastTableDWidth[bplvtotal].data[4]=FastTableDWidth[bplvtotal].data[5]=
            FastTableDWidth[bplvtotal].data[6]=FastTableDWidth[bplvtotal].data[7]=tmp;
            
            tmp=VideoULA_Palette[beebpixvr];
            if (tmp>7) {
                tmp&=7;
                if (VideoULA_ControlReg & 1) tmp^=7;
            };
            FastTableDWidth[bplvtotal].data[8]=FastTableDWidth[bplvtotal].data[9]=
            FastTableDWidth[bplvtotal].data[10]=FastTableDWidth[bplvtotal].data[11]=
            FastTableDWidth[bplvtotal].data[12]=FastTableDWidth[bplvtotal].data[13]=
            FastTableDWidth[bplvtotal].data[14]=FastTableDWidth[bplvtotal].data[15]=tmp;
        }; /* beebpixr */
    }; /* beebpixl */
} /* DoFastTable16XStep8 */

/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses bits 7,5,3,1 for the       */
/* palette address, the next uses 6,4,2,0, the next uses 5,3,1,H (H=High), then 5,2,0,H                        */
static void DoFastTable4(void)
{
    unsigned char tmp;
    unsigned long beebpixv,pentry;
    
    for(beebpixv=0;beebpixv<256;beebpixv++) {
        pentry=((beebpixv & 128)?8:0)
        | ((beebpixv & 32)?4:0)
        | ((beebpixv & 8)?2:0)
        | ((beebpixv & 2)?1:0);
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTable[beebpixv].data[0]=FastTable[beebpixv].data[1]=tmp;
        
        pentry=((beebpixv & 64)?8:0)
        | ((beebpixv & 16)?4:0)
        | ((beebpixv & 4)?2:0)
        | ((beebpixv & 1)?1:0);
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTable[beebpixv].data[2]=FastTable[beebpixv].data[3]=tmp;
        
        pentry=((beebpixv & 32)?8:0)
        | ((beebpixv & 8)?4:0)
        | ((beebpixv & 2)?2:0)
        | 1;
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTable[beebpixv].data[4]=FastTable[beebpixv].data[5]=tmp;
        pentry=((beebpixv & 16)?8:0)
        | ((beebpixv & 4)?4:0)
        | ((beebpixv & 1)?2:0)
        | 1;
        tmp=VideoULA_Palette[pentry];
        if (tmp>7)
        {
            tmp &= 7;
            if (VideoULA_ControlReg & 1)
            {
                tmp ^= 7;
            }
        }
        
        FastTable[beebpixv].data[6]=FastTable[beebpixv].data[7]=tmp;
    }; /* beebpixv */
} /* DoFastTable4 */

/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses bits 7,5,3,1 for the       */
/* palette address, the next uses 6,4,2,0, the next uses 5,3,1,H (H=High), then 5,2,0,H                        */
static void DoFastTable4XStep4(void)
{
    unsigned char tmp;
    unsigned long beebpixv,pentry;
    
    for(beebpixv=0;beebpixv<256;beebpixv++) {
        pentry=((beebpixv & 128)?8:0)
        | ((beebpixv & 32)?4:0)
        | ((beebpixv & 8)?2:0)
        | ((beebpixv & 2)?1:0);
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTableDWidth[beebpixv].data[0]=FastTableDWidth[beebpixv].data[1]=
        FastTableDWidth[beebpixv].data[2]=FastTableDWidth[beebpixv].data[3]=tmp;
        
        pentry=((beebpixv & 64)?8:0)
        | ((beebpixv & 16)?4:0)
        | ((beebpixv & 4)?2:0)
        | ((beebpixv & 1)?1:0);
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTableDWidth[beebpixv].data[4]=FastTableDWidth[beebpixv].data[5]=
        FastTableDWidth[beebpixv].data[6]=FastTableDWidth[beebpixv].data[7]=tmp;
        
        pentry=((beebpixv & 32)?8:0)
        | ((beebpixv & 8)?4:0)
        | ((beebpixv & 2)?2:0)
        | 1;
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTableDWidth[beebpixv].data[8]=FastTableDWidth[beebpixv].data[9]=
        FastTableDWidth[beebpixv].data[10]=FastTableDWidth[beebpixv].data[11]=tmp;
        pentry=((beebpixv & 16)?8:0)
        | ((beebpixv & 4)?4:0)
        | ((beebpixv & 1)?2:0)
        | 1;
        tmp=VideoULA_Palette[pentry];
        if (tmp>7) {
            tmp&=7;
            if (VideoULA_ControlReg & 1) tmp^=7;
        };
        FastTableDWidth[beebpixv].data[12]=FastTableDWidth[beebpixv].data[13]=
        FastTableDWidth[beebpixv].data[14]=FastTableDWidth[beebpixv].data[15]=tmp;
    }; /* beebpixv */
    
} /* DoFastTable4XStep4 */

/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses the same pattern as mode 1 */
/* all the way upto the 5th pixel which uses 31hh then 20hh and hten 1hhh then 0hhhh                           */
static void DoFastTable2(void)
{
    
    unsigned char tmp;
    unsigned long beebpixv,beebpixvt,pentry;
    int pix;
    
    for(beebpixv=0;beebpixv<256;beebpixv++) {
        beebpixvt=beebpixv;
        for(pix=0;pix<8;pix++) {
            pentry=((beebpixvt & 128)?8:0)
            | ((beebpixvt & 32)?4:0)
            | ((beebpixvt & 8)?2:0)
            | ((beebpixvt & 2)?1:0);
            beebpixvt<<=1;
            beebpixvt|=1;
            tmp=VideoULA_Palette[pentry];
            if (tmp>7) {
                tmp&=7;
                if (VideoULA_ControlReg & 1) tmp^=7;
            };
            FastTable[beebpixv].data[pix]=tmp;
        }; /* pix */
    }; /* beebpixv */
    
}; /* DoFastTable2 */

/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses the same pattern as mode 1 */
/* all the way upto the 5th pixel which uses 31hh then 20hh and hten 1hhh then 0hhhh                           */
static void DoFastTable2XStep2(void)
{
    unsigned char tmp;
    unsigned long beebpixv,beebpixvt,pentry;
    int pix;
    
    for(beebpixv=0;beebpixv<256;beebpixv++) {
        beebpixvt=beebpixv;
        for(pix=0;pix<8;pix++) {
            pentry=((beebpixvt & 128)?8:0)
            | ((beebpixvt & 32)?4:0)
            | ((beebpixvt & 8)?2:0)
            | ((beebpixvt & 2)?1:0);
            beebpixvt<<=1;
            beebpixvt|=1;
            tmp=VideoULA_Palette[pentry];
            if (tmp>7) {
                tmp&=7;
                if (VideoULA_ControlReg & 1) tmp^=7;
            };
            FastTableDWidth[beebpixv].data[pix*2]=FastTableDWidth[beebpixv].data[pix*2+1]=tmp;
        }; /* pix */
    }; /* beebpixv */
}; /* DoFastTable2XStep2 */

/*-------------------------------------------------------------------------------------------------------------*/
/* Check validity of fast table, and if invalid rebuild.
 The fast table accelerates the translation of beeb video memory
 values into X pixel values */
static void DoFastTable(void)
{
    /* if it's already OK then quit */
    if (FastTable_Valid)
    {
        return;
    }
    
#if 0
    if (!(CRTC_HorizontalDisplayed & 3))
    {
        LineRoutine = (VideoULA_ControlReg & 0x10) ? LowLevelDoScanLineNarrow : LowLevelDoScanLineWide;
    }
    else
    {
#endif
        LineRoutine = (VideoULA_ControlReg & 0x10) ? LowLevelDoScanLineNarrowNot4Bytes : LowLevelDoScanLineWideNot4Bytes;
#if 0
    }
#endif
    
    /* What happens next dpends on the number of colours */
    switch (NColsLookup[(VideoULA_ControlReg & 0x1c) >> 2])
    {
        case 2:
            if (VideoULA_ControlReg & 0x10)
            {
                DoFastTable2();
            }
            else
            {
                DoFastTable2XStep2();
            }
            
            FastTable_Valid = 1;
            break;
            
        case 4:
            if (VideoULA_ControlReg & 0x10)
            {
                DoFastTable4();
            }
            else
            {
                DoFastTable4XStep4();
            }
            
            FastTable_Valid = 1;
            break;
            
        case 16:
            if (VideoULA_ControlReg & 0x10)
            {
                DoFastTable16();
            }
            else
            {
                DoFastTable16XStep8();
            }
            
            FastTable_Valid = 1;
            break;
            
        default:
            break;
    } /* Colours/pixel switch */
} /* DoFastTable */

void LowLevelDoScanLine()
{
    /* Update acceleration tables */
    DoFastTable();
    
    if (FastTable_Valid)
    {
        LineRoutine();
    }
    
} /* LowLevelDoScanLine */


/*--------------------------------------------------------------------------*/
/* Scanline processing for modes with fast 6845 clock - i.e. narrow pixels  */
static void LowLevelDoScanLineNarrow()
{
#if 0
    unsigned char *CurrentPtr;
    int BytesToGo=CRTC_HorizontalDisplayed;
    EightUChars *vidPtr=mainWin->GetLinePtr(VideoState.PixmapLine);
    
    /* If the step is 4 then each byte corresponds to one entry in the fasttable
     and thus we can copy it really easily (and fast!) */
    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    
    /* This should help the compiler - it doesn't need to test for end of loop
     except every 4 entries */
    BytesToGo/=4;
    for(;BytesToGo;CurrentPtr+=32,BytesToGo--) {
        *(vidPtr++)=FastTable[*CurrentPtr];
        *(vidPtr++)=FastTable[*(CurrentPtr+8)];
        *(vidPtr++)=FastTable[*(CurrentPtr+16)];
        *(vidPtr++)=FastTable[*(CurrentPtr+24)];
    };
#endif
} /* LowLevelDoScanLineNarrow() */

/*--------------------------------------------------------------------------*/
/* Scanline processing for modes with fast 6845 clock - i.e. narrow pixels  */
/* This version handles screen modes where there is not a multiple of 4     */
/* bytes per scanline.                                                      */
static void LowLevelDoScanLineNarrowNot4Bytes()
{
    unsigned char *CurrentPtr;
    int BytesToGo = CRTC_HorizontalDisplayed;
    //    EightUChars *vidPtr=mainWin->GetLinePtr(VideoState.PixmapLine);
    unsigned char *vidPtr  = currentFrame->GetLinePtr(VideoState.PixmapLine * 2);
    unsigned char *vidPtr2 = currentFrame->GetLinePtr(VideoState.PixmapLine * 2 + 1);
    
    int i;
    
    
    /* If the step is 4 then each byte corresponds to one entry in the fasttable
     and thus we can copy it really easily (and fast!) */
    CurrentPtr=(unsigned char *)VideoState.DataPtr + VideoState.InCharLineUp;
    
    for(; BytesToGo; CurrentPtr += 8, BytesToGo--)
    {
        //        (vidPtr++)->eightbyte=FastTable[*CurrentPtr].eightbyte;
        for(i = 0; i < 8; i++)
        {
            *vidPtr++ = FastTable[*CurrentPtr].data[i];
            *vidPtr2++ = FastTable[*CurrentPtr].data[i];
        }
    }
} /* LowLevelDoScanLineNarrowNot4Bytes() */

/*-----------------------------------------------------------------------------*/
/* Scanline processing for the low clock rate modes                            */
static void LowLevelDoScanLineWide()
{
#if 0
    unsigned char *CurrentPtr;
    int BytesToGo = CRTC_HorizontalDisplayed;
    //    SixteenUChars *vidPtr=mainWin->GetLinePtr16(VideoState.PixmapLine);
    unsigned int *vidPtr = currentFrame->GetLinePtr(VideoState.PixmapLine);
    
    /* If the step is 4 then each byte corresponds to one entry in the fasttable
     and thus we can copy it really easily (and fast!) */
    CurrentPtr = (unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    
    /* This should help the compiler - it doesn't need to test for end of loop
     except every 4 entries */
    BytesToGo /= 4;
    for(;BytesToGo;CurrentPtr+=32,BytesToGo--)
    {
        *(vidPtr++)=FastTableDWidth[*CurrentPtr];
        *(vidPtr++)=FastTableDWidth[*(CurrentPtr+8)];
        *(vidPtr++)=FastTableDWidth[*(CurrentPtr+16)];
        *(vidPtr++)=FastTableDWidth[*(CurrentPtr+24)];
    }
#endif
} /* LowLevelDoScanLineWide */

/*-----------------------------------------------------------------------------*/
/* Scanline processing for the low clock rate modes                            */
/* This version handles cases where the screen width is not divisible by 4     */
static void LowLevelDoScanLineWideNot4Bytes()
{
    unsigned char *CurrentPtr;
    int BytesToGo = CRTC_HorizontalDisplayed;
    //    SixteenUChars *vidPtr=mainWin->GetLinePtr16(VideoState.PixmapLine);
    //    unsigned int *vidPtr = currentFrame->GetLinePtr(VideoState.PixmapLine);
    unsigned char *vidPtr  = currentFrame->GetLinePtr(VideoState.PixmapLine * 2);
    unsigned char *vidPtr2 = currentFrame->GetLinePtr(VideoState.PixmapLine * 2 + 1);
    int i;
    
    CurrentPtr = (unsigned char *)VideoState.DataPtr + VideoState.InCharLineUp;
    
    for( ;BytesToGo; CurrentPtr+=8, BytesToGo--)
    {
        for(i = 0; i < 16; i++)
        {
            *vidPtr++ = FastTableDWidth[*CurrentPtr].data[i];
            *vidPtr2++ = FastTableDWidth[*CurrentPtr].data[i];
            //        *(vidPtr++)=FastTableDWidth[*CurrentPtr];
        }
    }
} /* LowLevelDoScanLineWideNot4Bytes */




