//
//  BeebMemory.cpp
//  BeebEmX
//
//  Created by Steven Bagley on 16/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#include "BeebMemory.h"
#include "sysvia.h"

#include "FredJimMemory.h"
#include "SheilaMemory.h"

IMemory unreadable;

CBeebMemory::CBeebMemory() : m_osRom(NULL)
{
    /* Create default setup for a BBC Model B -- sub-classes can specialise this for a B+ or Master */
    int page;
    
    for(page = 0; page < 0x80; page++)
    {
        m_pageLUT[page] = &m_ram;
    }

    for(page = 0x80; page < 0x100; page++)
    {
        m_pageLUT[page] = &unreadable;
    }

    for(int i = 0; i < 16; i++)
    {
        m_roms[i] = NULL;
    }
    
    m_pageLUT[0xfc] = m_pageLUT[0xfd] = m_fred = new CFredJimMemory(this);
    m_pageLUT[0xfe] = m_sheila = new CSheilaMemory(this);
    
    /* Set up hardware */
    m_shadowDisplay = 0;
    m_bank = 0;
}


void CBeebMemory::RegisterDevice(int address, IMemory *device)
{
    m_sheila->RegisterDevice(address, device);
}

CBeebMemory::~CBeebMemory()
{
    int rom;
    for(rom = 0; rom < 16; rom++)
    {
        if(rom)
        {
            delete m_roms[rom];
        }
    }
    
    delete m_osRom;
}

void CBeebMemory::SetSWR(int bank)
{
    IMemory *pRom;
    
    m_bank = bank;
    pRom = m_roms[bank & 0xf];
    
    if(pRom == NULL)
    {
        pRom = &unreadable;
    }
    
    for(int page = 0x80; page < 0xc0; page++)
    {
        m_pageLUT[page] = pRom;
    }
    
}

void CBeebMemory::LoadOS(CBeebMemoryRom *osRom)
{
    if(m_osRom != NULL)
    {
        delete m_osRom;
    }
    
    m_osRom = osRom;
    
    for(int page = 0xc0; page < 0xfc; page++)
    {
        m_pageLUT[page] = m_osRom;
    }
    
    m_pageLUT[0xff] = m_osRom;
}

void CBeebMemory::LoadSWR(int pos, CBeebMemoryRom *rom)
{
    if(m_roms[pos] != NULL)
    {
        delete m_roms[pos];
    }
    
    m_roms[pos] = rom;
}

static unsigned int WrapAddr(int in)
{
    unsigned int offsets[]={0x4000,0x6000,0x3000,0x5800}; // page 419 of AUG is wrong
    if(in < 0x8000)
    {
        return(in);
    }
    
    in += offsets[(IC32State & 0x30) >> 4];
    in &= 0x7fff;
    return(in);
}


unsigned char *CBeebMemory::PtrWithWrap(int a, int n)
{
    static unsigned char tmpBuf[1024];
    unsigned char *tmpBufPtr;
    int EndAddr = a + n - 1;
    int toCopy;
    
    a = WrapAddr(a);
    EndAddr = WrapAddr(EndAddr);
    
#if 0
    // On Master the FSRam area is displayed if start addr below shadow area
    if(machine->MachineType() == 3 && a <= EndAddr && Sh_Display>0 && a < 0x3000)
    {
        if (0x3000-a < n)
        {
            toCopy=0x3000-a;
            if (toCopy>n) toCopy=n;
            if (toCopy>0) memcpy(tmpBuf,FSRam+0x2000-toCopy,toCopy);
            tmpBufPtr=tmpBuf+toCopy;
            toCopy=n-toCopy;
            if (toCopy>0) memcpy(tmpBufPtr,ShadowRAM+EndAddr-(toCopy-1),toCopy);
            return(tmpBuf);
        }
        else if (a<0x1000) {
            return((char *)FSRam); // Should probably be PrivateRAM?
        }
        else {
            return((char *)FSRam+a-0x1000);
        }
    }
#endif
    if (a <= EndAddr && m_shadowDisplay == 0)
    {
        return(m_ram.m_mainMemory+a);
    }
    
#if 0
    if (a<=EndAddr && Sh_Display > 0)
    {
        return((char *)ShadowRAM+a);
    }
#endif

    toCopy = 0x8000-a;
    if(toCopy > n)
    {
        toCopy = n;
    }
    
    if(toCopy > 0 && m_shadowDisplay == 0)
    {
        memcpy(tmpBuf, m_ram.m_mainMemory + a, toCopy);
    }
    
#if 0
    if (toCopy>0 && Sh_Display>0) memcpy(tmpBuf,ShadowRAM+a,toCopy);
#endif
    
    tmpBufPtr = tmpBuf + toCopy;
    toCopy = n - toCopy;

    if (toCopy>0 && m_shadowDisplay == 0)
    {
        memcpy(tmpBufPtr,m_ram.m_mainMemory+EndAddr-(toCopy-1),toCopy);
    }
    
#if 0
    if (toCopy>0 && Sh_Display>0) memcpy(tmpBufPtr,ShadowRAM+EndAddr-(toCopy-1),toCopy);
#endif
    // Tripling is for Shadow RAM handling
    return tmpBuf;
}

static unsigned int WrapAddrMo7(int in)
{
    if (in<0x8000)
    {
        return(in);
    }
    
    in+=0x7c00;
    in&=0x7fff;
    return(in);
}; /* WrapAddrMo7 */


unsigned char *CBeebMemory::PtrWithWrapMo7(int a, int n)
{
    static unsigned char tmpBuf[1024];
    unsigned char *tmpBufPtr;
    int EndAddr=a+n-1;
    int toCopy;
    
    a=WrapAddrMo7(a);
    EndAddr=WrapAddrMo7(EndAddr);
    
    if (a<=EndAddr && m_shadowDisplay==0)
    {
        return(m_ram.m_mainMemory + a);
    }
#if 0
    if (a<=EndAddr && m_shadowDisplay > 0)
    {
        return((char *)ShadowRAM+a);
    }
#endif
    
    toCopy=0x8000-a;
    if (toCopy>n && m_shadowDisplay == 0)
    {
        return(m_ram.m_mainMemory+a);
    }
    
#if 0
    if (toCopy>n && m_shadowDisplay > 0)
    {
        return((char *)ShadowRAM+a);
    }
#endif
    if (toCopy>0 && m_shadowDisplay == 0)
    {
        memcpy(tmpBuf,m_ram.m_mainMemory + a, toCopy);
    }
    
#if 0
    if (toCopy>0 && m_shadowDisplay > 0)
    {
        memcpy(tmpBuf,ShadowRAM+a,toCopy);
    }
#endif
    
    tmpBufPtr = tmpBuf + toCopy;
    toCopy = n - toCopy;
    
    if (toCopy>0 && m_shadowDisplay == 0)
    {
        memcpy(tmpBufPtr,m_ram.m_mainMemory + EndAddr - (toCopy - 1), toCopy);
    }
    
#if 0
    if (toCopy>0 && m_shadowDisplay > 0)
    {
        memcpy(tmpBufPtr,ShadowRAM+EndAddr-(toCopy-1),toCopy);
    }
#endif
    return tmpBuf;
}



