//
//  BeebMemory.h
//  BeebEmX
//
//  Created by Steven Bagley on 16/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#ifndef __BeebEmX__BeebMemory__
#define __BeebEmX__BeebMemory__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct IMemory
{
    virtual int ReadByte(const int address)
    {
        return 0xFF;
    }
    
    virtual void WriteByte(const int address, const unsigned char value)
    {
        return;
    }
    
};

class CSheilaMemory;
class CFredJimMemory;

class CBeebMemoryRom : public IMemory
{
public:
    CBeebMemoryRom(const void *rom, const int size)
    {
        memset(m_rom, 0, 16384);
        memcpy(m_rom, rom, size < 16384 ? size : 16384);
    }
    
    virtual int ReadByte(const int address)
    {
        return m_rom[address & 0x3fff];
    }

private:
    unsigned char m_rom[16384];
};



class CBeebMemory
{
public:
    CBeebMemory();
    ~CBeebMemory();
    
    inline int ReadByte(const int address)
    {
        return m_pageLUT[(address & 0xffff) >> 8]->ReadByte(address & 0xffff);

    }

    inline void WriteByte(const int address, const unsigned char value)
    {
        m_pageLUT[(address & 0xffff) >> 8]->WriteByte(address & 0xffff, value);
    }

    
    void RegisterDevice(int address, IMemory *device);

    
    void LoadOS(CBeebMemoryRom *osRom);
    void LoadSWR(int pos, CBeebMemoryRom *rom);
        
        
    
    void SetSWR(int bank);
    int GetSWR() { return m_bank; }
   
    /* Used to help video */
    unsigned char *PtrWithWrap(int a, int n);
    unsigned char *PtrWithWrapMo7(int a, int n);
private:
    
    class  : public IMemory
    {
    public:
        
        virtual int ReadByte(int address)
        {
            return m_mainMemory[address & 0x7fff];
        }
        
        virtual void WriteByte(const int address, const unsigned char value)
        {
            m_mainMemory[address & 0x7fff] = value;
        }
        unsigned char m_mainMemory[32768];
        
    } m_ram;
    

    IMemory *m_pageLUT[256];
    
    IMemory *m_roms[16];
    IMemory *m_osRom;
    CSheilaMemory *m_sheila;
    CFredJimMemory *m_fred;
    int m_shadowDisplay;
    int m_bank;
};



#endif /* defined(__BeebEmX__BeebMemory__) */
