//
//  TeletextVideo.cpp
//  BeebEmX
//
//  Created by Steven Bagley on 25/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#include "TeletextVideo.h"
#include "6502core.h"
#include "BeebMemory.h"
#include "VideoState.h"
#include "BeebFrame.h"

#include "sysvia.h"

extern int FrameNum;
extern int VScreenAdjust;
extern VideoStateT VideoState;
extern CBeebFrame *currentFrame;
char TeletextStyle = 1; // Defines wether teletext will skip intermediate lines in order to speed up
extern int ova, ovn; // mem ptr buffers

/* first subscript is graphics flag (1 for graphics,2 for separated graphics), next is character, then scanline */
/* character is (valu &127)-32 */
extern unsigned int EM7Font[3][96][20]; // 20 rows to account for "half pixels"

int Mode7FlashOn = 1; /* True if a flashing character in mode 7 is on */
static int Mode7DoubleHeightFlags[80]; /* Pessimistic size for this flags - if 1 then corresponding character on NEXT line is top half */
static int CurrentLineBottom = 0;
static int NextLineBottom = 0; // 1 if the next line of double height should be bottoms only


int Mode7FlashTrigger = MODE7ONFIELDS;




extern void VideoAddCursor(void);
extern void AdjustVideo();
extern void VideoAddLEDs(void);
extern void VideoStartOfFrame(void);

static void DoMode7Row(void);

void TeletextVideoDoScanLine(CBeebMemory *memory)
{
    static int DoCA1Int=0;
    int l;

    
    if (DoCA1Int)
    {
        SysVIATriggerCA1Int(0);
        DoCA1Int = 0;
    }
    
    /* Clear the next 20 scan lines */
    if (!FrameNum)
    {
        if(VScreenAdjust > 0 && VideoState.PixmapLine == 0)
        {
            for (l = -VScreenAdjust; l < 0; ++l)
            {
                currentFrame->doHorizLine(0, l, -36, CBeebFrame::kFrameWidth);
            }
        }
        
        for(l = 0; (l < 20) && ((VideoState.PixmapLine + l) < 512); ++l)
        {
            currentFrame->doHorizLine(0, VideoState.PixmapLine + l, -36, CBeebFrame::kFrameWidth);
        }
    }
    
    // RTW - Mode 7 emulation is rather broken, as we should be plotting it line-by-line instead
    // of character block at a time.
    // For now though, I leave it as it is, and plot an entire block when InCharLineUp is 0.
    // The infrastructure now exists though to make DoMode7Row plot just a single scanline (from InCharLineUp 0..9).
    if (VideoState.CharLine < CRTC_VerticalDisplayed && VideoState.InCharLineUp==0)
    {
        ova = VideoState.Addr;
        ovn = CRTC_HorizontalDisplayed;
        VideoState.DataPtr = memory->PtrWithWrapMo7(VideoState.Addr, CRTC_HorizontalDisplayed);
        VideoState.Addr += CRTC_HorizontalDisplayed;
        
        if (!FrameNum)
        {
            DoMode7Row();
        }
        
        VideoState.PixmapLine += 20;
    };
    
    /* Move onto next physical scanline as far as timing is concerned */
    VideoState.InCharLineUp += 1;
    
    // RTW - Mode 7 hardwired for now. Assume 10 scanlines per character regardless (actually 9.5 but god knows how that works)
    if(VideoState.CharLine <= CRTC_VerticalTotal && VideoState.InCharLineUp > 9)
    {
        VideoState.CharLine++;
        VideoState.InCharLineUp = 0;
    }
    
    // RTW - check if we have reached the end of the PAL frame.
    // This whole thing is a bit hardwired and kludged. Should really be emulating VSync position 'properly' like Modes 0-6.
    if (VideoState.CharLine > CRTC_VerticalTotal && VideoState.InCharLineUp >= CRTC_VerticalTotalAdjust)
    {
        // Changed so that whole screen is still visible after *TV255
        VScreenAdjust = -100 + (((CRTC_VerticalTotal+1)-(CRTC_VerticalSyncPos-1))*(20/TeletextStyle));
        AdjustVideo();
        if (!FrameNum)
        {
            VideoAddCursor();
            VideoAddLEDs();
            // Clear rest of screen below virtical total
            for (l = VideoState.PixmapLine; l < 500 / TeletextStyle; ++l)
            {
                currentFrame->doHorizLine(0, l, -36, CBeebFrame::kFrameWidth);
            }
        }
        VideoState.IsNewTVFrame = 1;
        VideoStartOfFrame();
        VideoState.PreviousLastPixmapLine = VideoState.PixmapLine;
        VideoState.PixmapLine = 0;
        SysVIATriggerCA1Int(1);
        DoCA1Int = 1;
    }
    else
    {
        // RTW- set timer till the next scanline update (this is now nice and simple)
        IncTrigger((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2),VideoTriggerCount);
    }
}


/*-------------------------------------------------------------------------------------------------------------*/
/* Do all the pixel rows for one row of teletext characters                                                    */
static void DoMode7Row(void)
{
    unsigned char *CurrentPtr = VideoState.DataPtr;
    int CurrentChar;
    int XStep;
    unsigned char byte;
    unsigned int tmp;
    
    unsigned int Foreground = 7;
    unsigned int ActualForeground;
    unsigned int Background = 0;
    int Flash = 0; /* i.e. steady */
    int DoubleHeight = 0; /* Normal */
    int Graphics = 0; /* I.e. alpha */
    int Separated = 0; /* i.e. continuous graphics */
    int HoldGraph = 0; /* I.e. don't hold graphics - I don't know what hold graphics is anyway! */
    // That's ok. Nobody else does either, and nor do I. - Richard Gellman.
    int HoldGraphChar = 32; // AHA! we know what it is now, this is the character to "hold" during control codes
    unsigned int CurrentCol[20]={0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff
        ,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff};
    int CurrentLen[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    int CurrentStartX[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    int CurrentScanLine;
    int CurrentX = 0;
    int CurrentPixel;
    unsigned int col;
    int FontTypeIndex = 0; /* 0=alpha, 1=contiguous graphics, 2=separated graphics */
    
    if (CRTC_HorizontalDisplayed > 80)
    {
        return; /* Not possible on beeb - and would break the double height lookup array */
    }
    
    XStep=1;
    
    for(CurrentChar = 0; CurrentChar < CRTC_HorizontalDisplayed; CurrentChar++)
    {
        byte = CurrentPtr[CurrentChar];
        if(byte < 32)
        {
            byte+=128; // fix for naughty programs that use 7-bit control codes - Richard Gellman
        }
        
        if((byte & 32) && (Graphics))
        {
            HoldGraphChar = byte;
        }
        
        if((byte>=128) && (byte<=159))
        {
            switch(byte)
            {
                case 129: case 130: case 131: case 132:
                case 133: case 134: case 135:
                    Foreground = byte - 128;
                    Graphics = 0;
                    break;
                    
                case 136:
                    Flash = 1;
                    break;
                    
                case 137:
                    Flash = 0;
                    break;
                    
                case 140:
                    DoubleHeight = 0;
                    break;
                    
                case 141:
                    if(!CurrentLineBottom)
                    {
                        NextLineBottom=1;
                    }
                    DoubleHeight = 1;
                    break;
                    
                case 145: case 146: case 147: case 148:
                case 149: case 150: case 151:
                    Foreground = byte-144;
                    Graphics = 1;
                    break;
                    
                case 152: /* Conceal display - not sure about this */
                    Foreground = Background;
                    break;
                    
                case 153:
                    Separated = 0;
                    break;
                    
                case 154:
                    Separated = 1;
                    break;
                    
                case 156:
                    Background = 0;
                    break;
                    
                case 157:
                    Background = Foreground;
                    break;
                    
                case 158:
                    HoldGraph = 1;
                    break;
                    
                case 159:
                    HoldGraph = 0;
                    break;
            } /* Special character switch */
            
            // This next line hides any non double height characters on the bottom line
            /* Fudge so that the special character is just displayed in background */
            if((HoldGraph == 1) && (Graphics == 1))
            {
                byte=HoldGraphChar;
            }
            else
            {
                byte=32;
            }
            FontTypeIndex = Graphics ? (Separated ? 2 : 1) : 0;
        } /* test for special character */
        
        if((CurrentLineBottom) && ((byte & 127) > 31) && (!DoubleHeight))
        {
            byte = 32;
        }
        
        if ((CRTC_ScanLinesPerChar <= 9) || (THalfMode))
        {
            TeletextStyle = 2;
        }
        else
        {
            TeletextStyle = 1;
        }
        
        /* Top bit never reaches character generator */
        byte &= 127;
        
        /* Our font table goes from character 32 up */
        if (byte < 32)
        {
            byte = 0;
        }
        else
        {
            byte -= 32;
        }
        
        /* Conceal flashed text if necessary */
        ActualForeground = (Flash && !Mode7FlashOn) ? Background : Foreground;
        if(!DoubleHeight)
        {
            for(CurrentScanLine = 0 + (TeletextStyle - 1); CurrentScanLine < 20; CurrentScanLine += TeletextStyle)
            {
                tmp = EM7Font[FontTypeIndex][byte][CurrentScanLine];
                
                //tmp=1365;
                
                if ((tmp == 0) || (tmp == 255))
                {
                    col = (tmp == 0) ? Background : ActualForeground;
                    
                    if (col == CurrentCol[CurrentScanLine])
                    {
                        CurrentLen[CurrentScanLine] += 12 * XStep;
                    }
                    else
                    {
                        if(CurrentLen[CurrentScanLine])
                        {
                            currentFrame->doHorizLine(CurrentCol[CurrentScanLine], VideoState.PixmapLine + CurrentScanLine, CurrentStartX[CurrentScanLine], CurrentLen[CurrentScanLine]);
                        }
                        
                        CurrentCol[CurrentScanLine] = col;
                        CurrentStartX[CurrentScanLine] = CurrentX;
                        CurrentLen[CurrentScanLine]=12*XStep;
                        
                    } /* same colour */
                }
                else
                {
                    for(CurrentPixel = 0x800; CurrentPixel; CurrentPixel = CurrentPixel>>1)
                    {
                        /* Background or foreground ? */
                        col = (tmp & CurrentPixel) ? ActualForeground : Background;
                        
                        /* Do we need to draw ? */
                        if (col == CurrentCol[CurrentScanLine])
                        {
                            CurrentLen[CurrentScanLine] += XStep;
                        }
                        else
                        {
                            if (CurrentLen[CurrentScanLine])
                            {
                                currentFrame->doHorizLine(CurrentCol[CurrentScanLine], VideoState.PixmapLine + CurrentScanLine, CurrentStartX[CurrentScanLine], CurrentLen[CurrentScanLine]);
                            }
                            CurrentCol[CurrentScanLine] = col;
                            CurrentStartX[CurrentScanLine] = CurrentX;
                            CurrentLen[CurrentScanLine] = XStep;
                        } /* Fore/back ground */
                        
                        CurrentX+=XStep;
                    } /* Pixel within byte */
                    
                    CurrentX -= 12*XStep;
                } /* tmp!=0 */
            } /* Scanline for */
            
            CurrentX += 12 * XStep;
            Mode7DoubleHeightFlags[CurrentChar] = 1; /* Not double height - so if the next line is double height it will be top half */
        }
        else /* Double height! */
        {
            int ActualScanLine;
            
            for(CurrentPixel=0x800; CurrentPixel; CurrentPixel = CurrentPixel >> 1)
            {
                for(CurrentScanLine = 0 + (TeletextStyle - 1); CurrentScanLine < 20; CurrentScanLine += TeletextStyle)
                {
                    if(!CurrentLineBottom)
                    {
                        ActualScanLine = CurrentScanLine >> 1;
                    }
                    else
                    {
                        ActualScanLine=10+(CurrentScanLine>>1);
                    }
                    
                    /* Background or foreground ? */
                    col=(EM7Font[FontTypeIndex][byte][ActualScanLine] & CurrentPixel) ? ActualForeground : Background;
                    
                    /* Do we need to draw ? */
                    if (col==CurrentCol[CurrentScanLine])
                    {
                        CurrentLen[CurrentScanLine]+=XStep;
                    }
                    else
                    {
                        if (CurrentLen[CurrentScanLine])
                        {
                            currentFrame->doHorizLine(CurrentCol[CurrentScanLine], VideoState.PixmapLine + CurrentScanLine, CurrentStartX[CurrentScanLine], CurrentLen[CurrentScanLine]);
                        }
                        CurrentCol[CurrentScanLine]=col;
                        CurrentStartX[CurrentScanLine]=CurrentX;
                        CurrentLen[CurrentScanLine]=XStep;
                    } /* Fore/back ground */
                } /* Scanline for */
                
                CurrentX += XStep;
            } /* Pixel within byte */
            Mode7DoubleHeightFlags[CurrentChar] ^= 1; /* Not double height - so if the next line is double height it will be top half */
        }
    } /* character loop */
    
    /* Finish off right bits of scan line */
    for(CurrentScanLine = 0 + (TeletextStyle - 1); CurrentScanLine < 20; CurrentScanLine += TeletextStyle)
    {
        if (CurrentLen[CurrentScanLine])
        {
            currentFrame->doHorizLine(CurrentCol[CurrentScanLine], VideoState.PixmapLine + CurrentScanLine, CurrentStartX[CurrentScanLine], CurrentLen[CurrentScanLine]);
        }
    }
    CurrentLineBottom = NextLineBottom;
    NextLineBottom = 0;
} /* DoMode7Row */




