/****************************************************************
BeebEm - BBC Micro and Master 128 Emulator
Copyright (C) 1994  Nigel Magnay
Copyright (C) 1997  Mike Wyatt
Copyright (C) 1998  Robert Schmidt
Copyright (C) 2001  Richard Gellman
Copyright (C) 2004  Ken Lowe
Copyright (C) 2004  Rob O'Donnell
Copyright (C) 2005  Jon Welch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public 
License along with this program; if not, write to the Free 
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
****************************************************************/

// 07/06/1997: Mike Wyatt and NRM's port to Win32
// 11/01/1998: Conveted to use DirectX, Mike Wyatt
// 28/12/2004: Econet added Rob O'Donnell. robert@irrelevant.com.
// 26/12/2011: Added IDE Drive to Hardware options, JGH

#include <stdio.h>
#include <unistd.h>
#include "beebwin.h"
#include "6502core.h"
#include "disc8271.h"
#include "beebmem.h"
#include "beebsound.h"
#include "serial.h"
#include "tube.h"
#include "z80mem.h"
#include "z80.h"
#include "Arm.h"
#include "sysvia.h"
#include "uservia.h"
#include "video.h"
#include "disc1770.h"

#include "serialdevices.h"
#if 0
#include <windows.h>
#include <initguid.h>
#include <shlobj.h>
#include <shlwapi.h>
#include <gdiplus.h>
#include "main.h"
#include "port.h"
#include "beebemrc.h"
#include "atodconv.h"
#include "userkybd.h"
#include "econet.h"	 // Rob O'Donnell Christmas 2004.
#include "ext1770.h"
#include "uefstate.h"
#include "debug.h"
#include "scsi.h"
#include "sasi.h"
#include "ide.h"
#include "userkybd.h"
#ifdef SPEECH_ENABLED
#include "speech.h"
#endif
#include "teletext.h"
#include "avi.h"
#include "csw.h"
#include "version.h"
#endif



unsigned char MachineType;

#if 0
#define WIN_STYLE (WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_MAXIMIZEBOX|WS_SIZEBOX)

// some LED based macros
#define LED_COL_BASE 64

static char *CFG_REG_KEY = "Software\\BeebEm";



#ifdef M512COPRO_ENABLED
void i86_main(void);
#endif

#endif
FILE *CMDF2;
unsigned char CMA2;
CArm *arm = NULL;

struct LEDType LEDs;
char DiscLedColour=0; // 0 for red, 1 for green.
#if 0
unsigned char HideMenuEnabled;
unsigned char DisableMenu = 0;
bool MenuOn = true;


AVIWriter *aviWriter = NULL;

// FDC Board extension DLL variables
HMODULE hFDCBoard;

EDCB ExtBoard={0,0,NULL};
bool DiscLoaded[2]={FALSE,FALSE}; // Set to TRUE when a disc image has been loaded.
char CDiscName[2][256]; // Filename of disc current in drive 0 and 1;
char CDiscType[2]; // Current disc types
char FDCDLL[256]={0};

const char *WindowTitle = "BeebEm - BBC Model B / Master 128 Emulator";
static const char *AboutText =
	"BeebEm - Emulating:\n\nBBC Micro Model B\nBBC Micro Model B + IntegraB\n"
	"BBC Micro Model B Plus (128)\nAcorn Master 128\n\n"
	"Acorn 65C02 Second Processor\n"
	"Torch Z80 Second Processor\nAcorn Z80 Second Processor\n"
#ifdef M512COPRO_ENABLED
	"Master 512 Second Processor\n"
#endif
	"ARM Second Processor\n\n"
	"Version " VERSION_STRING ", Feb 2012";

/* Prototypes */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

#endif
// Keyboard mappings
static KeyMap defaultMapping;
static KeyMap logicalMapping;

/* Currently selected translation table */
static KeyMap *transTable = &defaultMapping;

/****************************************************************************/
BeebWin::BeebWin()
{   
	m_LastStartY = 0;
	m_LastNLines = 256;
#if 0
    m_LastTickCount = 0;
#endif
	m_KeyMapAS = 0;
	m_KeyMapFunc = 0;
	m_ShiftPressed = 0;
	m_ShiftBooted = false;
	for (int k = 0; k < 256; ++k)
	{
		m_vkeyPressed[k][0][0] = -1;
		m_vkeyPressed[k][1][0] = -1;
		m_vkeyPressed[k][0][1] = -1;
		m_vkeyPressed[k][1][1] = -1;
	}
	m_DisableKeysWindows=0;
	m_DisableKeysBreak=0;
	m_DisableKeysEscape=0;
	m_DisableKeysShortcut=0;
	memset(&defaultMapping, 0, sizeof(KeyMap));
	memset(&logicalMapping, 0, sizeof(KeyMap));
#if 0
	memset(&UserKeymap, 0, sizeof(KeyMap));
	memset(m_UserKeyMapPath, 0, sizeof(m_UserKeyMapPath));
#endif
	m_frozen = false;
#if 0
    IgnoreIllegalInstructions = 1;
#endif
	for(int i=0;i < 256;i++)
		cols[i] = i;
#if 0
	m_WriteProtectDisc[0] = !IsDiscWritable(0);
	m_WriteProtectDisc[1] = !IsDiscWritable(1);
    UEFTapeName[0]=0;
#endif
	m_AutoSavePrefsCMOS = false;
	m_AutoSavePrefsFolders = false;
	m_AutoSavePrefsAll = false;
	m_AutoSavePrefsChanged = false;
	memset(m_KbdCmd, 0, sizeof(m_KbdCmd));
	memset(m_DebugScript, 0, sizeof(m_DebugScript));
	m_KbdCmdPos = -1;
	m_KbdCmdPress = false;
	m_KbdCmdDelay = 40;
	m_KbdCmdLastCycles = 0;
	m_NoAutoBoot = false;
	m_clipboardlen = 0;
	m_clipboardptr = 0;
	m_printerbufferlen = 0;
	m_translateCRLF = true;
	m_CurrentDisplayRenderer = 0;
#if 0
	m_DXSmoothing = true;
	m_DXSmoothMode7Only = false;
	m_DXResetPending = false;
	m_JoystickCaptured = false;
#endif
    m_customip[0] = 0;
	m_customport = 0;
	m_isFullScreen = false;
	m_MaintainAspectRatio = true;
	m_startFullScreen = false;
#if 0
    m_XDXSize = 640;
	m_YDXSize = 480;
	m_XScrSize = GetSystemMetrics(SM_CXSCREEN);
	m_YScrSize = GetSystemMetrics(SM_CYSCREEN);
	m_XWinBorder = GetSystemMetrics(SM_CXSIZEFRAME) * 2;
	m_YWinBorder = GetSystemMetrics(SM_CYSIZEFRAME) * 2 +
		GetSystemMetrics(SM_CYMENUSIZE) +
		GetSystemMetrics(SM_CYCAPTION) + 1;
#endif
    
	/* Get the applications path - used for non-user files */
#if 0
    char app_path[_MAX_PATH];
	char app_drive[_MAX_DRIVE];
	char app_dir[_MAX_DIR];
	GetModuleFileName(NULL, app_path, _MAX_PATH);
	_splitpath(app_path, app_drive, app_dir, NULL, NULL);
	_makepath(m_AppPath, app_drive, app_dir, NULL, NULL);
    
	// Read user data path from registry
	if (!RegGetStringValue(HKEY_CURRENT_USER, CFG_REG_KEY, "UserDataFolder", m_UserDataPath, _MAX_PATH))
	{
		// Default user data path to a sub-directory in My Docs
		if (SHGetFolderPath(NULL, CSIDL_PERSONAL, NULL, SHGFP_TYPE_CURRENT, m_UserDataPath) == NOERROR)
		{
			strcat(m_UserDataPath, "\\BeebEm\\");
		}
	}

	// Read disc images path from registry
	if (!RegGetStringValue(HKEY_CURRENT_USER, CFG_REG_KEY, "DiscsPath", m_DiscPath, _MAX_PATH))
	{
		// Default disc images path to a sub-directory of UserData path
		strcpy(m_DiscPath, m_UserDataPath);
		strcat(m_DiscPath, "DiscIms\\");
	}

#endif
    // Set default files, may be overridden by command line parameters.
	strcpy(m_PrefsFile, "Preferences.cfg");
#if 0
	strcpy(RomFile, "Roms.cfg");
#endif
}

/****************************************************************************/
void BeebWin::Initialise(EMVIDEOLINK *vlink)
{   
	// Parse command line
	ParseCommandLine();
	FindCommandLineFile(m_CommandLineFileName1);
	FindCommandLineFile(m_CommandLineFileName2);
	CheckForLocalPrefs(m_CommandLineFileName1, false);

	// Check that user data directory exists
	if (CheckUserDataPath() == false)
		exit(-1);
    
    TubeEnabled = 0;
    MachineType = 0;
    VideoLink = vlink;
#if 0
	LoadPreferences();
#endif
	// Override full screen?
	if (m_startFullScreen)
	{
		m_isFullScreen = true;
	}

    CreateBeebWindow();
	CreateBitmap();

#if 0
    m_hMenu = GetMenu(m_hWnd);
	m_hDC = GetDC(m_hWnd);
#endif

#if 0
	ReadROMFile(RomFile, RomConfig);
#endif
    ApplyPrefs();

#if 0
    if(strcmp(m_DebugScript,"\0") != 0)
	{
		DebugOpenDialog(hInst, m_hWnd);
		DebugRunScript(m_DebugScript);
	}
#endif
	// Boot file if passed on command line
	HandleCommandLineFile(1, m_CommandLineFileName2);
	HandleCommandLineFile(0, m_CommandLineFileName1);

#if 0
    // Schedule first key press if keyboard command supplied
	if (m_KbdCmd[0] != 0)
		SetTimer(m_hWnd, 1, 1000, NULL);
#endif
}

/****************************************************************************/
void BeebWin::ApplyPrefs()
{
#if 0
	// Set up paths
	strcpy(EconetCfgPath, m_UserDataPath);
	strcpy(RomPath, m_UserDataPath);
	strcpy(DiscPath, m_DiscPath);

	// Load key maps
	char keymap[_MAX_PATH];
	strcpy(keymap, "Logical.kmap");
	GetDataPath(m_UserDataPath, keymap);
	ReadKeyMap(keymap, &logicalMapping);
	strcpy(keymap, "Default.kmap");
	GetDataPath(m_UserDataPath, keymap);
	ReadKeyMap(keymap, &defaultMapping);

	InitMenu();
	ShowMenu(TRUE);

	ExitDX();
	if (m_DisplayRenderer != IDM_DISPGDI)
		InitDX();

	InitTextToSpeech();
	InitTextView();

	/* Initialise printer */
	if (PrinterEnabled)
		PrinterEnable(m_PrinterDevice);
	else
		PrinterDisable();

	/* Joysticks can only be initialised after the window is created (needs hwnd) */
	if (m_MenuIdSticks == IDM_JOYSTICK)
		InitJoystick();

	LoadFDC(NULL, true);
	RTCInit();

#if 0
	SoundReset();
	if(SoundDefault)
    {
        SoundInit();
    }
	SetSoundMenu();
#endif
    
#ifdef SPEECH_ENABLED
	if (SpeechDefault)
		tms5220_start();
#endif

	// Serial init
	if (IP232custom)
	{
		strcpy(IPAddress,IP232customip);
		PortNo = IP232customport;
	}
	else
	{
		strcpy(IPAddress,"127.0.0.1");
		PortNo = 25232;
	}
	if (SerialPortEnabled) {
		if (TouchScreenEnabled) 
			TouchScreenOpen();

		if (EthernetPortEnabled && (IP232localhost || IP232custom))
		{
			if (IP232Open() == false)
			{
				MessageBox(GETHWND,"Serial IP232 could not connect to specified address",
						   WindowTitle,MB_OK|MB_ICONERROR);
				bSerialStateChanged=TRUE;
				SerialPortEnabled=FALSE;
				UpdateSerialMenu(m_hMenu);
			}
		}
	}
    
	ResetBeebSystem(MachineType,TubeEnabled/*,1*/);

	// Rom write flags
	for (int slot = 0; slot < 16; ++slot)
	{
		if (!RomWritePrefs[slot])
			RomWritable[slot] = 0;
	}
	SetRomMenu();
    
#endif
}

/****************************************************************************/
BeebWin::~BeebWin()
{   
#if 0
    if (m_DisplayRenderer != IDM_DISPGDI)
		ExitDX();

	ReleaseDC(m_hWnd, m_hDC);

	if (m_hOldObj != NULL)
		SelectObject(m_hDCBitmap, m_hOldObj);
	if (m_hBitmap != NULL)
		DeleteObject(m_hBitmap);
	if (m_hDCBitmap != NULL)
		DeleteDC(m_hDCBitmap);

	GdiplusShutdown(m_gdiplusToken);

	CoUninitialize();
#endif
}

/****************************************************************************/
void BeebWin::Shutdown()
{
#if 0
	if (aviWriter)
		delete aviWriter;
#endif 
    
#if 0
	if (m_AutoSavePrefsCMOS || m_AutoSavePrefsFolders || m_AutoSavePrefsAll || m_AutoSavePrefsChanged)
		SavePreferences(m_AutoSavePrefsAll);

    if (SoundEnabled)
		SoundReset();
#endif

#if 0
	if (m_SpVoice)
	{
		m_SpVoice->Release();
		m_SpVoice = NULL;
	}
#endif
    
#if 0
    /* SRB FIX ME */
	if (mEthernetHandle > 0)
	{
		IP232Close();
		mEthernetHandle = 0;
	}
#endif
}


void BeebWin::InitialiseMemory()
{
    int bank;
    
    BeebMemInit(false,m_ShiftBooted);

    // Clear ROMs
    for (bank = 0; bank < 16; bank++)
    {
        RomWritable[bank] = 0;
        RomBankType[bank] = BankEmpty;
        memset(Roms[bank], 0, 0x4000);
    }
    
}



void BeebWin::Tick()
{
    Exec6502Instruction();
}

int  BeebWin::CyclesExecuted()
{
    return TotalCycles;
}

void BeebWin::LoadMode7Font(const unsigned char *fontData)
{
    BuildMode7Font(fontData);
}


void BeebWin::Set1770DFS()
{
    NativeFDC = false;
    EFDCAddr = 0xfe84;
    EDCAddr = 0xfe80;
    InvertTR00 = false;
}

void BeebWin::LoadDisc(char *path, int DscDrive, int DscType)
{
    Load1770DiscImage(path, DscDrive, DscType /*,HMENU dmenu*/);
}


void BeebWin::LoadRom(int romPosition, const void *data, unsigned int size)
{
    if(romPosition >= 0 && romPosition < 16)
    {
        RomWritable[romPosition] = 0;
        RomBankType[romPosition] = BankRom;
        
        /* Copy ROM image */
        memcpy(Roms[romPosition], data, size < 16384 ? size : 16384);
    }
}

void BeebWin::LoadOSRom(const void *data, unsigned int size)
{
    memcpy(WholeRam+0xC000, data, size < 16384 ? size : 16384);
    
    
}

/****************************************************************************/
void BeebWin::ResetBeebSystem(unsigned char NewModelType, unsigned char TubeStatus/*, unsigned char LoadRoms*/)
{
	SwitchOnCycles = 0; // Reset delay
#if 0
    SoundReset();
	if (SoundDefault)
    {
		SoundInit();
    }
	SwitchOnSound();
#endif

    
#if 0
    EnableTube = TubeStatus;
#endif
	MachineType = NewModelType;

    
    /* SRB HACK */
    
    /* Put first ROM in */
    memcpy(WholeRam + 0x8000, Roms[0xf], 0x4000);
    PagedRomReg=0xf;
    
	Init6502core();
#if 0
	if (EnableTube) Init65C02core();
#endif
#ifdef M512COPRO_ENABLED
	if (Tube186Enabled) i86_main();
#endif
#if 0
    Enable_Z80 = 0;
    if (TorchTube || AcornZ80)
    {
       R1Status = 0;
       ResetTube();
       init_z80();
       Enable_Z80 = 1;
    }
	Enable_Arm = 0;
	if (ArmTube)
	{
		R1Status = 0;
		ResetTube();
		if (arm) delete arm;
		arm = new CArm;
		Enable_Arm = 1;
	}
#endif
	SysVIAReset();
	UserVIAReset();
	VideoInit();
    Reset1770();
#if 0
    SetDiscWriteProtects();
	Disc8271_reset();
	if (EconetEnabled) EconetReset();	//Rob:
	AtoDInit();
	SetRomMenu();
	FreeDiscImage(0);
	// Keep the disc images loaded
	FreeDiscImage(1);
	Close1770Disc(0);
	Close1770Disc(1);
	if (SCSIDriveEnabled) SCSIReset();
	if (SCSIDriveEnabled) SASIReset();
	if (IDEDriveEnabled)  IDEReset();
	TeleTextInit();
	if (MachineType==3) InvertTR00=FALSE;
	if (MachineType!=3) {
		LoadFDC(NULL, false);
	}
	if ((MachineType!=3) && (NativeFDC)) {
		// 8271 disc
		if ((DiscLoaded[0]) && (CDiscType[0]==0)) LoadSimpleDiscImage(CDiscName[0],0,0,80);
		if ((DiscLoaded[0]) && (CDiscType[0]==1)) LoadSimpleDSDiscImage(CDiscName[0],0,80);
		if ((DiscLoaded[1]) && (CDiscType[1]==0)) LoadSimpleDiscImage(CDiscName[1],1,0,80);
		if ((DiscLoaded[1]) && (CDiscType[1]==1)) LoadSimpleDSDiscImage(CDiscName[1],1,80);
	}
	if (((MachineType!=3) && (!NativeFDC)) || (MachineType==3)) {
		// 1770 Disc
		if (DiscLoaded[0]) Load1770DiscImage(CDiscName[0],0,CDiscType[0],m_hMenu);
		if (DiscLoaded[1]) Load1770DiscImage(CDiscName[1],1,CDiscType[1],m_hMenu);
	}
#endif
}

/****************************************************************************/
void BeebWin::CreateBitmap()
{
#if 0
	if (m_hBitmap != NULL)
		DeleteObject(m_hBitmap);
	if (m_hDCBitmap != NULL)
		DeleteDC(m_hDCBitmap);
	if (m_screen_blur != NULL)
		free(m_screen_blur);

	m_hDCBitmap = CreateCompatibleDC(NULL);

	m_bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	m_bmi.bmiHeader.biWidth = 800;
	m_bmi.bmiHeader.biHeight = -512;
	m_bmi.bmiHeader.biPlanes = 1;
	m_bmi.bmiHeader.biBitCount = 8;
	m_bmi.bmiHeader.biXPelsPerMeter = 0;
	m_bmi.bmiHeader.biYPelsPerMeter = 0;
	m_bmi.bmiHeader.biCompression = BI_RGB;
	m_bmi.bmiHeader.biSizeImage = 800*512;
	m_bmi.bmiHeader.biClrUsed = 68;
	m_bmi.bmiHeader.biClrImportant = 68;

#ifdef USE_PALETTE
	__int16 *pInts = (__int16 *)&m_bmi.bmiColors[0];
    
	for(int i=0; i<12; i++)
		pInts[i] = i;
    
	m_hBitmap = CreateDIBSection(m_hDCBitmap, (BITMAPINFO *)&m_bmi, DIB_PAL_COLORS,
							(void**)&m_screen, NULL,0);
#else
	for (int i = 0; i < 64; ++i)
	{
		float r,g,b;
		r = (float) (i & 1);
		g = (float) ((i & 2) >> 1);
		b = (float) ((i & 4) >> 2);

		if (palette_type != RGB)
		{
			r = g = b = (float) (0.299 * r + 0.587 * g + 0.114 * b);

			switch (palette_type)
			{
			case AMBER:
				r *= (float) 1.0;
				g *= (float) 0.8;
				b *= (float) 0.1;
				break;
			case GREEN:
				r *= (float) 0.2;
				g *= (float) 0.9;
				b *= (float) 0.1;
				break;
			}
		}

		m_bmi.bmiColors[i].rgbRed   = (BYTE) (r * m_BlurIntensities[i >> 3] / 100.0 * 255);
		m_bmi.bmiColors[i].rgbGreen = (BYTE) (g * m_BlurIntensities[i >> 3] / 100.0 * 255);
		m_bmi.bmiColors[i].rgbBlue  = (BYTE) (b * m_BlurIntensities[i >> 3] / 100.0 * 255);
		m_bmi.bmiColors[i].rgbReserved = 0;
	}

	// Red Leds - left is dark, right is lit.
	m_bmi.bmiColors[LED_COL_BASE].rgbRed=80;		m_bmi.bmiColors[LED_COL_BASE+1].rgbRed=255;
	m_bmi.bmiColors[LED_COL_BASE].rgbGreen=0;		m_bmi.bmiColors[LED_COL_BASE+1].rgbGreen=0;
	m_bmi.bmiColors[LED_COL_BASE].rgbBlue=0;		m_bmi.bmiColors[LED_COL_BASE+1].rgbBlue=0;
	m_bmi.bmiColors[LED_COL_BASE].rgbReserved=0;	m_bmi.bmiColors[LED_COL_BASE+1].rgbReserved=0;
	// Green Leds - left is dark, right is lit.
	m_bmi.bmiColors[LED_COL_BASE+2].rgbRed=0;		m_bmi.bmiColors[LED_COL_BASE+3].rgbRed=0;
	m_bmi.bmiColors[LED_COL_BASE+2].rgbGreen=80;	m_bmi.bmiColors[LED_COL_BASE+3].rgbGreen=255;
	m_bmi.bmiColors[LED_COL_BASE+2].rgbBlue=0;		m_bmi.bmiColors[LED_COL_BASE+3].rgbBlue=0;
	m_bmi.bmiColors[LED_COL_BASE+2].rgbReserved=0;	m_bmi.bmiColors[LED_COL_BASE+3].rgbReserved=0;

	m_hBitmap = CreateDIBSection(m_hDCBitmap, (BITMAPINFO *)&m_bmi, DIB_RGB_COLORS,
							(void**)&m_screen, NULL,0);
#endif

	m_screen_blur = (char *)calloc(m_bmi.bmiHeader.biSizeImage,1);

	m_hOldObj = SelectObject(m_hDCBitmap, m_hBitmap);
	if(m_hOldObj == NULL)
		MessageBox(m_hWnd,"Cannot select the screen bitmap\n"
					"Try running in a 256 colour mode",WindowTitle,MB_OK|MB_ICONERROR);
#endif
}


/****************************************************************************/
void BeebWin::CreateBeebWindow(void)
{
#if 0
	DWORD style;
	int x,y;
	int show = SW_SHOW;

	x = m_XWinPos;
	y = m_YWinPos;
	if (x == -1 || y == -1)
	{
		x = CW_USEDEFAULT;
		y = 0;
		m_XWinPos = 0;
		m_YWinPos = 0;
	}

	if (m_DisplayRenderer == IDM_DISPGDI && m_isFullScreen)
		show = SW_MAXIMIZE;

	if (m_DisplayRenderer != IDM_DISPGDI && m_isFullScreen)
	{
		style = WS_POPUP;
	}
	else
	{
		style = WIN_STYLE;
	}

	m_hWnd = CreateWindow(
				"BEEBWIN",				// See RegisterClass() call.
				m_szTitle, 		// Text for window title bar.
				style,
				x, y,
				m_XWinSize + m_XWinBorder,
				m_YWinSize + m_YWinBorder,
				NULL,					// Overlapped windows have no parent.
				NULL,				 // Use the window class menu.
				hInst,			 // This instance owns this window.
				NULL				 // We don't use any data in our WM_CREATE
		); 

	ShowWindow(m_hWnd, show); // Show the window
	UpdateWindow(m_hWnd);		  // Sends WM_PAINT message

	SetWindowAttributes(false);
#endif
}

/****************************************************************************/
void BeebWin::SetRomMenu(void)
{
#if 0
	HMENU hMenu = m_hMenu;

	// Set the ROM Titles in the ROM/RAM menu.
	CHAR Title[19];
	
	int	 i;

	for( i=0; i<16; i++ )
	{
		Title[0] = '&';
		_itoa( i, &Title[1], 16 );
		Title[2] = ' ';
		
		// Get the Rom Title.
		ReadRomTitle( i, &Title[3], sizeof( Title )-4);
	
		if ( Title[3]== '\0' )
		{
			if (RomBankType[i]==BankRam)
				strcpy( &Title[3], "RAM" );
			else
				strcpy( &Title[3], "Empty" );
		}

		ModifyMenu( hMenu,	// handle of menu 
					IDM_ALLOWWRITES_ROM0 + i,
					MF_BYCOMMAND,	// menu item to modify
				//	MF_STRING,	// menu item flags 
					IDM_ALLOWWRITES_ROM0 + i,	// menu item identifier or pop-up menu handle
					Title		// menu item content 
					);

		/* Disable ROM and uncheck the Rom/RAM which are NOT writable */
		EnableMenuItem(hMenu, IDM_ALLOWWRITES_ROM0 + i,
					   (RomBankType[i]==BankRam) ? MF_ENABLED : MF_GRAYED);
		CheckMenuItem(hMenu, IDM_ALLOWWRITES_ROM0 + i,
					  (RomBankType[i]==BankRam) ? (RomWritable[i] ? MF_CHECKED : MF_UNCHECKED) : MF_UNCHECKED);
	}
#endif
}

/****************************************************************************/
void BeebWin::InitJoystick(void)
{
#if 0
	MMRESULT mmresult = JOYERR_NOERROR;

	if (!m_JoystickCaptured)
	{
		/* Get joystick updates 10 times a second */
		mmresult = joySetCapture(m_hWnd, JOYSTICKID1, 100, FALSE);
		if (mmresult == JOYERR_NOERROR)
			mmresult = joyGetDevCaps(JOYSTICKID1, &m_JoystickCaps, sizeof(JOYCAPS));
		if (mmresult == JOYERR_NOERROR)
			m_JoystickCaptured = true;
	}

	if (mmresult == JOYERR_NOERROR)
	{
		AtoDEnable();
	}
	else if (mmresult == JOYERR_UNPLUGGED)
	{
		MessageBox(m_hWnd, "Joystick is not plugged in",
					WindowTitle, MB_OK|MB_ICONERROR);
	}
	else
	{
		MessageBox(m_hWnd, "Failed to initialise the joystick",
					WindowTitle, MB_OK|MB_ICONERROR);
	}
#endif
}

/****************************************************************************/
void BeebWin::ScaleJoystick(unsigned int x, unsigned int y)
{
#if 0
	if (m_MenuIdSticks == IDM_JOYSTICK)
	{
		/* Scale and reverse the readings */
		JoystickX = (int)((double)(m_JoystickCaps.wXmax - x) * 65535.0 /
						  (double)(m_JoystickCaps.wXmax - m_JoystickCaps.wXmin));
		JoystickY = (int)((double)(m_JoystickCaps.wYmax - y) * 65535.0 /
						  (double)(m_JoystickCaps.wYmax - m_JoystickCaps.wYmin));
	}
#endif
}

/****************************************************************************/
void BeebWin::ResetJoystick(void)
{
#if 0
	// joySetCapture() fails after a joyReleaseCapture() call (not sure why)
	// so leave joystick captured.
	// joyReleaseCapture(JOYSTICKID1);
	AtoDDisable();
#endif
}

/****************************************************************************/
void BeebWin::SetMousestickButton(int button)
{
#if 0
	if ( (m_MenuIdSticks == IDM_AMOUSESTICK) || ((m_MenuIdSticks == IDM_DMOUSESTICK)) )
		JoystickButton = button;
#endif
}

/****************************************************************************/
void BeebWin::ScaleMousestick(unsigned int x, unsigned int y)
{
#if 0
	static int lastx = 32768;
	static int lasty = 32768;
	int dx, dy;

	if (m_MenuIdSticks == IDM_AMOUSESTICK)				// Analogue mouse stick
	{
		JoystickX = (m_XWinSize - x) * 65535 / m_XWinSize;
		JoystickY = (m_YWinSize - y) * 65535 / m_YWinSize;
	}
	else if (m_MenuIdSticks == IDM_DMOUSESTICK)		// Digital mouse stick
	{
		dx = x - lastx;
		dy = y - lasty;
		
		if (dx > 4) JoystickX = 0;
		if (dx < -4) JoystickX = 65535;

		if (dy > 4) JoystickY = 0;
		if (dy < -4) JoystickY = 65535;

		lastx = x;
		lasty = y;
	}

	if (m_HideCursor)
		SetCursor(NULL);
#endif
}

/****************************************************************************/
void BeebWin::SetAMXPosition(unsigned int x, unsigned int y)
{
#if 0
	if (AMXMouseEnabled)
	{
		// Scale the window coords to the beeb screen coords
		AMXTargetX = x * m_AMXXSize * (100 + m_AMXAdjust) / 100 / m_XWinSize;
		AMXTargetY = y * m_AMXYSize * (100 + m_AMXAdjust) / 100 / m_YWinSize;

		AMXMouseMovement();
	}
#endif
}


void BeebWin::KeyDown(int row, int col)
{
    BeebKeyDown(row, col);
}

void BeebWin::KeyUp(int row, int col)
{
    BeebKeyUp(row, col);
}


/****************************************************************************/
int BeebWin::TranslateKey(int vkey, int keyUp, int &row, int &col)
{
	if (vkey < 0 || vkey > 255)
		return -9;

	// Key track of shift state
	if ((*transTable)[vkey][0].row == 0 && (*transTable)[vkey][0].col == 0)
	{
		if (keyUp)
			m_ShiftPressed = 0;
		else
			m_ShiftPressed = 1;
	}

	if (keyUp)
	{
		// Key released, lookup beeb row + col that this vkey 
		// mapped to when it was pressed.  Need to release
		// both shifted and non-shifted presses.
		row = m_vkeyPressed[vkey][0][0];
		col = m_vkeyPressed[vkey][1][0];
		m_vkeyPressed[vkey][0][0] = -1;
		m_vkeyPressed[vkey][1][0] = -1;
		if (row >= 0)
			BeebKeyUp(row, col);

		row = m_vkeyPressed[vkey][0][1];
		col = m_vkeyPressed[vkey][1][1];
		m_vkeyPressed[vkey][0][1] = -1;
		m_vkeyPressed[vkey][1][1] = -1;
		if (row >= 0)
			BeebKeyUp(row, col);
	}
	else // New key press - convert to beeb row + col
	{
		int needShift = m_ShiftPressed;

		row = (*transTable)[vkey][m_ShiftPressed].row;
		col = (*transTable)[vkey][m_ShiftPressed].col;
		needShift = (*transTable)[vkey][m_ShiftPressed].shift;

		if (m_KeyMapAS)
		{
			// Map A & S to CAPS & CTRL - good for some games
			if (vkey == 65)
			{
				row = 4;
				col = 0;
			}
			else if (vkey == 83)
			{
				row = 0;
				col = 1;
			}
		}

		if (m_KeyMapFunc)
		{
			// Map F1-F10 to f0-f9
			if (vkey >= 113 && vkey <= 121)
			{
				row = (*transTable)[vkey - 1][0].row;
				col = (*transTable)[vkey - 1][0].col;
			}
			else if (vkey == 112)
			{
				row = 2;
				col = 0;
			}
		}

		if (m_DisableKeysShortcut && (row == -3 || row == -4))
			row = -10;
		if (m_DisableKeysEscape && row == 7 && col == 0)
			row = -10;
		if (m_DisableKeysBreak && row == -2 && col == -2)
			row = -10;

		if (row >= 0)
		{
			// Make sure shift state is correct
			if (needShift)
				BeebKeyDown(0, 0);
			else
				BeebKeyUp(0, 0);

			BeebKeyDown(row, col);

			// Record beeb row + col for key release
			m_vkeyPressed[vkey][0][m_ShiftPressed ? 1 : 0] = row;
			m_vkeyPressed[vkey][1][m_ShiftPressed ? 1 : 0] = col;
		}
		else
		{
			// Special key!  Record so key up returns correct codes
			m_vkeyPressed[vkey][0][1] = row;
			m_vkeyPressed[vkey][1][1] = col;
		}
	}

	return(row);
}

/****************************************************************************/
int BeebWin::StartOfFrame(void)
{
	int FrameNum = 1;

	if (UpdateTiming())
		FrameNum = 0;
#if 0
	// Force video frame rate to match AVI capture rate to avoid 
	// video and sound getting out of sync
	if (aviWriter != NULL)
	{
		m_AviFrameSkipCount++;
		if (m_AviFrameSkipCount > m_AviFrameSkip)
		{
			m_AviFrameSkipCount = 0;
			FrameNum = 0;
		}
		else
		{
			FrameNum = 1;
		}

		// Ensure that frames captured each second (50 frames) matches
		// the AVI capture FPS rate
		m_AviFrameCount++;
		if (m_AviFrameCount >= 50)
		{
			m_AviFrameCount = 0;
			m_AviFrameSkipCount = 0;
		}
	}
#endif
	return FrameNum;
}

void BeebWin::doLED(int sx,bool on)
{
#if 0
	int tsy; char colbase;
	colbase=(DiscLedColour*2)+LED_COL_BASE; // colour will be 0 for red, 1 for green.
	if (sx<100) colbase=LED_COL_BASE; // Red leds for keyboard always
	if (TeletextEnabled)
		tsy=496;
	else
		tsy=m_LastStartY+m_LastNLines-2;
	doUHorizLine(mainWin->cols[((on)?1:0)+colbase],tsy,sx,8);
	doUHorizLine(mainWin->cols[((on)?1:0)+colbase],tsy,sx,8);
#endif
};

/****************************************************************************/
void BeebWin::ResetTiming(void)
{
#if 0
	m_LastTickCount = GetTickCount();
	m_LastStatsTickCount = m_LastTickCount;
	m_LastTotalCycles = TotalCycles;
	m_LastStatsTotalCycles = TotalCycles;
	m_TickBase = m_LastTickCount;
	m_CycleBase = TotalCycles;
	m_MinFrameCount = 0;
	m_LastFPSCount = m_LastTickCount;
	m_ScreenRefreshCount = 0;
#endif
}

/****************************************************************************/
bool BeebWin::UpdateTiming(void)
{
#if 0
	DWORD TickCount;
	DWORD Ticks;
	DWORD SpareTicks;

	int Cycles;
	int CyclesPerSec;
#endif

	bool UpdateScreen = FALSE;
#if 0
	TickCount = GetTickCount();

	/* Don't do anything if this is the first call or there has
	   been a long pause due to menu commands, or when something
	   wraps. */
	if (m_LastTickCount == 0 || TickCount < m_LastTickCount || (TickCount - m_LastTickCount) > 1000 || TotalCycles < m_LastTotalCycles)
	{
		ResetTiming();
		return TRUE;
	}

	/* Update stats every second */
	if (TickCount >= m_LastStatsTickCount + 1000)
	{
		m_FramesPerSecond = m_ScreenRefreshCount;
		m_ScreenRefreshCount = 0;
		m_RelativeSpeed = ((TotalCycles - m_LastStatsTotalCycles) / 2000.0) / (TickCount - m_LastStatsTickCount);
		m_LastStatsTotalCycles = TotalCycles;
		m_LastStatsTickCount += 1000;
		DisplayTiming();
	}

	// Now we work out if BeebEm is running too fast or not
	if (m_RealTimeTarget > 0.0)
	{
		Ticks = TickCount - m_TickBase;
		Cycles = (int)((double)(TotalCycles - m_CycleBase) / m_RealTimeTarget);

		if (Ticks <= (DWORD)(Cycles / 2000))
		{
			// Need to slow down, show frame (max 50fps though) 
			// and sleep a bit
			if (TickCount >= m_LastFPSCount + 20)
			{
				UpdateScreen = TRUE;
				m_LastFPSCount += 20;
			}
			else
			{
				UpdateScreen = FALSE;
			}

			SpareTicks = (DWORD)(Cycles / 2000) - Ticks;
			Sleep(SpareTicks);
			m_MinFrameCount = 0;
		}
		else
		{
			// Need to speed up, skip a frame
			UpdateScreen = FALSE;

			// Make sure we show at least one in 100 frames
			++m_MinFrameCount;
			if (m_MinFrameCount >= 100)
			{
				UpdateScreen = TRUE;
				m_MinFrameCount = 0;
			}
		}

		/* Move counter bases forward */
		CyclesPerSec = (int) (2000000.0 * m_RealTimeTarget);
		while ((TickCount - m_TickBase) > 1000 && (TotalCycles - m_CycleBase) > CyclesPerSec)
		{
			m_TickBase += 1000;
			m_CycleBase += CyclesPerSec;
		}
	}
	else
	{
		/* Fast as possible with a certain frame rate */
		if (TickCount >= m_LastFPSCount + (1000 / m_FPSTarget))
		{
			UpdateScreen = TRUE;
			m_LastFPSCount += 1000 / m_FPSTarget;
		}
		else
		{
			UpdateScreen = FALSE;
		}
	}

	m_LastTickCount = TickCount;
	m_LastTotalCycles = TotalCycles;
#endif
	return UpdateScreen;
}

/****************************************************************************/
void BeebWin::TranslateDDSize(void)
{
#if 0
	switch (m_DDFullScreenMode)
	{
	default:
	case ID_VIEW_DD_640X480:
		m_XDXSize = 640;
		m_YDXSize = 480;
		break;
	case ID_VIEW_DD_720X576:
		m_XDXSize = 720;
		m_YDXSize = 576;
		break;
	case ID_VIEW_DD_800X600:
		m_XDXSize = 800;
		m_YDXSize = 600;
		break;
	case ID_VIEW_DD_1024X768:
		m_XDXSize = 1024;
		m_YDXSize = 768;
		break;
	case ID_VIEW_DD_1280X720:
		m_XDXSize = 1280;
		m_YDXSize = 720;
		break;
	case ID_VIEW_DD_1280X768:
		m_XDXSize = 1280;
		m_YDXSize = 768;
		break;
	case ID_VIEW_DD_1280X960:
		m_XDXSize = 1280;
		m_YDXSize = 960;
		break;
	case ID_VIEW_DD_1280X1024:
		m_XDXSize = 1280;
		m_YDXSize = 1024;
		break;
	case ID_VIEW_DD_1440X900:
		m_XDXSize = 1440;
		m_YDXSize = 900;
		break;
	case ID_VIEW_DD_1600X1200:
		m_XDXSize = 1600;
		m_YDXSize = 1200;
		break;
	case ID_VIEW_DD_1920X1080:
		m_XDXSize = 1920;
		m_YDXSize = 1080;
		break;
	}
#endif
}
	
/****************************************************************************/
void BeebWin::TranslateWindowSize(void)
{
#if 0
	switch (m_MenuIdWinSize)
	{
	case IDM_320X256:
		m_XWinSize = 320;
		m_YWinSize = 256;
		break;
	default:
	case IDM_640X512:
		m_XWinSize = 640;
		m_YWinSize = 512;
		break;
	case IDM_800X600:
		m_XWinSize = 800;
		m_YWinSize = 600;
		break;
	case IDM_1024X512:
		m_XWinSize = 1024;
		m_YWinSize = 512;
		break;
	case IDM_1024X768:
		m_XWinSize = 1024;
		m_YWinSize = 768;
		break;
	case IDM_1280X1024:
		m_XWinSize = 1280;
		m_YWinSize = 1024;
		break;
	case IDM_1440X1080:
		m_XWinSize = 1440;
		m_YWinSize = 1080;
		break;
	case IDM_1600X1200:
		m_XWinSize = 1600;
		m_YWinSize = 1200;
		break;
	case IDM_CUSTOMWINSIZE:
		break;
	}

	m_XLastWinSize = m_XWinSize;
	m_YLastWinSize = m_YWinSize;

	m_MenuIdWinSize = IDM_CUSTOMWINSIZE;
#endif
}

/****************************************************************************/
void BeebWin::TranslateSampleRate(void)
{
#if 0
	switch (m_MenuIdSampleRate)
	{
	case IDM_44100KHZ:
		SoundSampleRate = 44100;
		break;

	default:
	case IDM_22050KHZ:
		SoundSampleRate = 22050;
		break;

	case IDM_11025KHZ:
		SoundSampleRate = 11025;
		break;
	}
#endif
}

/****************************************************************************/
void BeebWin::TranslateVolume(void)
{
#if 0
	switch (m_MenuIdVolume)
	{
	case IDM_FULLVOLUME:
		SoundVolume = 1;
		break;

	case IDM_HIGHVOLUME:
		SoundVolume = 2;
		break;

	default:
	case IDM_MEDIUMVOLUME:
		SoundVolume = 3;
		break;

	case IDM_LOWVOLUME:
		SoundVolume = 4;
		break;
	}
#endif
}

/****************************************************************************/
void BeebWin::TranslateTiming(void)
{
#if 0
	m_FPSTarget = 0;
	m_RealTimeTarget = 1.0;

	if (m_MenuIdTiming == IDM_3QSPEED)
		m_MenuIdTiming = IDM_FIXEDSPEED0_75;
	if (m_MenuIdTiming == IDM_HALFSPEED)
		m_MenuIdTiming = IDM_FIXEDSPEED0_5;

	switch (m_MenuIdTiming)
	{
	default:
	case IDM_REALTIME:
		m_RealTimeTarget = 1.0;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED100:
		m_RealTimeTarget = 100.0;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED50:
		m_RealTimeTarget = 50.0;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED10:
		m_RealTimeTarget = 10.0;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED5:
		m_RealTimeTarget = 5.0;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED2:
		m_RealTimeTarget = 2.0;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED1_5:
		m_RealTimeTarget = 1.5;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED1_25:
		m_RealTimeTarget = 1.25;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED1_1:
		m_RealTimeTarget = 1.1;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED0_9:
		m_RealTimeTarget = 0.9;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED0_5:
		m_RealTimeTarget = 0.5;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED0_75:
		m_RealTimeTarget = 0.75;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED0_25:
		m_RealTimeTarget = 0.25;
		m_FPSTarget = 0;
		break;

	case IDM_FIXEDSPEED0_1:
		m_RealTimeTarget = 0.1;
		m_FPSTarget = 0;
		break;

	case IDM_50FPS:
		m_FPSTarget = 50;
		m_RealTimeTarget = 0;
		break;

	case IDM_25FPS:
		m_FPSTarget = 25;
		m_RealTimeTarget = 0;
		break;

	case IDM_10FPS:
		m_FPSTarget = 10;
		m_RealTimeTarget = 0;
		break;

	case IDM_5FPS:
		m_FPSTarget = 5;
		m_RealTimeTarget = 0;
		break;

	case IDM_1FPS:
		m_FPSTarget = 1;
		m_RealTimeTarget = 0;
		break;
	}
#endif
	ResetTiming();
}

void BeebWin::AdjustSpeed(bool up)
{
#if 0
	static int speeds[] = {
				IDM_FIXEDSPEED100,
				IDM_FIXEDSPEED50,
				IDM_FIXEDSPEED10,
				IDM_FIXEDSPEED5,
				IDM_FIXEDSPEED2,
				IDM_FIXEDSPEED1_5,
				IDM_FIXEDSPEED1_25,
				IDM_FIXEDSPEED1_1,
				IDM_REALTIME,
				IDM_FIXEDSPEED0_9,
				IDM_FIXEDSPEED0_75,
				IDM_FIXEDSPEED0_5,
				IDM_FIXEDSPEED0_25,
				IDM_FIXEDSPEED0_1,
				0};
	int s = 0;
	int t = m_MenuIdTiming;

	while (speeds[s] != 0 && speeds[s] != m_MenuIdTiming)
		s++;

	if (speeds[s] == 0)
	{
		t = IDM_REALTIME;
	}
	else if (up)
	{
		if (s > 0)
			t = speeds[s-1];
	}
	else
	{
		if (speeds[s+1] != 0)
			t = speeds[s+1];
	}

	if (t != m_MenuIdTiming)
	{
		CheckMenuItem(m_hMenu, m_MenuIdTiming, MF_UNCHECKED);
		m_MenuIdTiming = t;
		CheckMenuItem(m_hMenu, m_MenuIdTiming, MF_CHECKED);
		TranslateTiming();
	}
#endif
}

/****************************************************************************/
void BeebWin::TranslateKeyMapping(void)
{
#if 0
	switch (m_MenuIdKeyMapping)
	{
	default:
	case IDM_DEFAULTKYBDMAPPING:
		transTable = &defaultMapping;
		break;

	case IDM_LOGICALKYBDMAPPING:
		transTable = &logicalMapping;
		break;

	case IDM_USERKYBDMAPPING:
		transTable = &UserKeymap;
		break;
	}
#endif
}

/****************************************************************************/
void BeebWin::SetTapeSpeedMenu(void)
{
#if 0
	CheckMenuItem(m_hMenu,ID_TAPE_FAST,(TapeClockSpeed==750)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_TAPE_MFAST,(TapeClockSpeed==1600)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_TAPE_MSLOW,(TapeClockSpeed==3200)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_TAPE_NORMAL,(TapeClockSpeed==5600)?MF_CHECKED:MF_UNCHECKED);
#endif
}

/****************************************************************************/
#define ASPECT_RATIO_X 5
#define ASPECT_RATIO_Y 4
void BeebWin::CalcAspectRatioAdjustment(int DisplayWidth, int DisplayHeight)
{
	m_XRatioAdj = 0.0f;
	m_YRatioAdj = 0.0f;
	m_XRatioCrop = 0.0f;
	m_YRatioCrop = 0.0f;

	if (m_isFullScreen)
	{
		int w = DisplayWidth * ASPECT_RATIO_Y;
		int h = DisplayHeight * ASPECT_RATIO_X;
		if (w > h)
		{
			m_XRatioAdj = (float)DisplayHeight / (float)DisplayWidth * (float)ASPECT_RATIO_X/(float)ASPECT_RATIO_Y;
			m_XRatioCrop = (1.0f - m_XRatioAdj) / 2.0f;
		}
		else if (w < h)
		{
			m_YRatioAdj = (float)DisplayWidth / (float)DisplayHeight * (float)ASPECT_RATIO_Y/(float)ASPECT_RATIO_X;
			m_YRatioCrop = (1.0f - m_YRatioAdj) / 2.0f;
		}
	}
}

/****************************************************************************/
void BeebWin::TranslateAMX(void)
{
#if 0
	switch (m_MenuIdAMXSize)
	{
	case IDM_AMX_160X256:
		m_AMXXSize = 160;
		m_AMXYSize = 256;
		break;
	default:
	case IDM_AMX_320X256:
		m_AMXXSize = 320;
		m_AMXYSize = 256;
		break;
	case IDM_AMX_640X256:
		m_AMXXSize = 640;
		m_AMXYSize = 256;
		break;
	}

	switch (m_MenuIdAMXAdjust)
	{
	case 0:
		m_AMXAdjust = 0;
		break;
	case IDM_AMX_ADJUSTP50:
		m_AMXAdjust = 50;
		break;
	default:
	case IDM_AMX_ADJUSTP30:
		m_AMXAdjust = 30;
		break;
	case IDM_AMX_ADJUSTP10:
		m_AMXAdjust = 10;
		break;
	case IDM_AMX_ADJUSTM10:
		m_AMXAdjust = -10;
		break;
	case IDM_AMX_ADJUSTM30:
		m_AMXAdjust = -30;
		break;
	case IDM_AMX_ADJUSTM50:
		m_AMXAdjust = -50;
		break;
	}
#endif
}


void BeebWin::UpdateOptiMenu(void)
{
#if 0
	CheckMenuItem(m_hMenu,ID_DOCONLY,(OpCodes==1)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_EXTRAS ,(OpCodes==2)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_FULLSET,(OpCodes==3)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_BHARDWARE,(BHardware==1)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_TSTYLE,(THalfMode==1)?MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(m_hMenu,ID_PSAMPLES,(PartSamples)?MF_CHECKED:MF_UNCHECKED);
#endif
}

void BeebWin::SetPBuff(void)
{
#if 0
	CheckMenuItem(m_hMenu,ID_PBUFF,(UsePrimaryBuffer)?MF_CHECKED:MF_UNCHECKED);
#endif
}

void BeebWin::SetSoundMenu(void)
{
#if 0
	CheckMenuItem(m_hMenu,IDM_SOUNDONOFF,(SoundEnabled && SoundDefault)?MF_CHECKED:MF_UNCHECKED);
	SetPBuff();
#endif
}

void BeebWin::Activate(bool active)
{
    
	if (active)
		m_frozen = FALSE;
	else if (m_FreezeWhenInactive)
		m_frozen = TRUE;
}

bool BeebWin::IsFrozen(void)
{
	return m_frozen;
}

/*****************************************************************************/
// Parse command line parameters

void BeebWin::ParseCommandLine()
{
	char errstr[200];
	int i;
	int a;
	bool invalid;

#if 0
	m_CommandLineFileName1[0] = 0;
	m_CommandLineFileName2[0] = 0;

	i = 1;
	while (i < __argc)
	{
		// Params with no arguments
		if (_stricmp(__argv[i], "-DisMenu") == 0)
		{
			DisableMenu = 1;
		}
		else if (_stricmp(__argv[i], "-NoAutoBoot") == 0)
		{
			m_NoAutoBoot = true;
		}
		else if (_stricmp(__argv[i], "-FullScreen") == 0)
		{
			m_startFullScreen = true;
		}
		else if (__argv[i][0] == '-' && i+1 >= __argc)
		{
			sprintf(errstr,"Invalid command line parameter:\n  %s", __argv[i]);
			MessageBox(m_hWnd,errstr,WindowTitle,MB_OK|MB_ICONERROR);
		}
		else // Params with additional arguments
		{
			invalid = false;
			if (_stricmp(__argv[i], "-Data") == 0)
			{
				strcpy(m_UserDataPath, __argv[++i]);

				if (strcmp(m_UserDataPath, "-") == 0)
				{
					// Use app path
					strcpy(m_UserDataPath, m_AppPath);
					strcat(m_UserDataPath, "UserData\\");
				}
				else
				{
					if (m_UserDataPath[strlen(m_UserDataPath)-1] != '\\' &&
						m_UserDataPath[strlen(m_UserDataPath)-1] != '/')
					{
						strcat(m_UserDataPath, "\\");
					}
				}
			}
			else if (_stricmp(__argv[i], "-Prefs") == 0)
			{
				strcpy(m_PrefsFile, __argv[++i]);
			}
			else if (_stricmp(__argv[i], "-Roms") == 0)
			{
				strcpy(RomFile, __argv[++i]);
			}
			else if (_stricmp(__argv[i], "-EcoStn") == 0)
			{
				a = atoi(__argv[++i]);
				if (a < 1 || a > 254)
					invalid = true;
				else
					EconetStationNumber = a;
			}
			else if (_stricmp(__argv[i], "-EcoFF") == 0)
			{
				a = atoi(__argv[++i]);
				if (a < 1)
					invalid = true;
				else
					EconetFlagFillTimeout = a;
			}
			else if (_stricmp(__argv[i], "-KbdCmd") == 0)
			{
				strncpy(m_KbdCmd, __argv[++i], sizeof(m_KbdCmd));
			}
			else if (_stricmp(__argv[i], "-DebugScript") == 0)
			{
				strncpy(m_DebugScript, __argv[++i], sizeof(m_DebugScript));
			}
			else if (__argv[i][0] == '-')
			{
				invalid = true;
				++i;
			}
			else
			{
				// Assume its a file name
				if (m_CommandLineFileName1[0] == 0)
					strncpy(m_CommandLineFileName1, __argv[i], _MAX_PATH);
				else if (m_CommandLineFileName2[0] == 0)
					strncpy(m_CommandLineFileName2, __argv[i], _MAX_PATH);
			}

			if (invalid)
			{
				sprintf(errstr,"Invalid command line parameter:\n  %s %s", __argv[i-1], __argv[i]);
				MessageBox(m_hWnd,errstr,WindowTitle,MB_OK|MB_ICONERROR);
			}
		}

		++i;
	}
#endif
}

/*****************************************************************************/
// Check for preference files in the same directory as the file specified
void BeebWin::CheckForLocalPrefs(const char *path, bool bLoadPrefs)
{
#if 0
	char file[_MAX_PATH];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	FILE *fd;

	if (path[0] == 0)
		return;

	_splitpath(path, drive, dir, NULL, NULL);

	// Look for prefs file
	_makepath(file, drive, dir, "Preferences", "cfg");
	fd = fopen(file, "r");
	if (fd != NULL)
	{
		fclose(fd);
		// File exists, use it
		strcpy(m_PrefsFile, file);
		if (bLoadPrefs)
		{
			LoadPreferences();

			// Reinit with new prefs
			SetWindowPos(m_hWnd, HWND_TOP, m_XWinPos, m_YWinPos,
						 0, 0, SWP_NOSIZE);
			HandleCommand(m_DisplayRenderer);
			InitMenu();
			SetWindowText(m_hWnd, WindowTitle);
			if (m_MenuIdSticks == IDM_JOYSTICK)
				InitJoystick();
		}
	}

	// Look for ROMs file
	_makepath(file, drive, dir, "Roms", "cfg");
	fd = fopen(file, "r");
	if (fd != NULL)
	{
		fclose(fd);
		// File exists, use it
		strcpy(RomFile, file);
		if (bLoadPrefs)
		{
			ReadROMFile(RomFile, RomConfig);
			BeebReadRoms();
		}
	}
#endif
}

/*****************************************************************************/
// File location of a file passed on command line
	
void BeebWin::FindCommandLineFile(char *CmdLineFile)
{
#if 0
	bool ssd = false;
	bool dsd = false;
	bool adfs = false;
	bool cont = false;
	char TmpPath[_MAX_PATH];
	char *FileName = NULL;
	bool uef = false;
	bool csw = false;
	bool img = false;
	
	// See if file is readable
	if (CmdLineFile[0] != 0)
	{
		FileName = CmdLineFile;
		strncpy(TmpPath, CmdLineFile, _MAX_PATH);

		// Work out which type of files it is
		char *ext = strrchr(FileName, '.');
		if (ext != NULL)
		{
			cont = true;
			if (_stricmp(ext+1, "ssd") == 0)
				ssd = true;
			else if (_stricmp(ext+1, "dsd") == 0)
				dsd = true;
			else if (_stricmp(ext+1, "adl") == 0)
				adfs = true;
			else if (_stricmp(ext+1, "adf") == 0)
				adfs = true;
			else if (_stricmp(ext+1, "uef") == 0)
				uef = true;
			else if (_stricmp(ext+1, "csw") == 0)
				csw = true;
			else if (_stricmp(ext+1, "img") == 0)
				img = true;
			else
			{
				char errstr[200];
				sprintf(errstr,"Unrecognised file type:\n  %s", FileName);
				MessageBox(m_hWnd,errstr,WindowTitle,MB_OK|MB_ICONERROR);
				cont = false;
			}
		}
	}

	if (cont)
	{
		cont = false;

		FILE *fd = fopen(FileName, "rb");
		if (fd != NULL)
		{
			cont = true;
			fclose(fd);
		}
		else if (uef)
		{
			// Try getting it from BeebState directory
			strcpy(TmpPath, m_UserDataPath);
			strcat(TmpPath, "beebstate/");
			strcat(TmpPath, FileName);
			FILE *fd = fopen(TmpPath, "rb");
			if (fd != NULL)
			{
				cont = true;
				FileName = TmpPath;
				fclose(fd);
			}
			else
			{
				// Try getting it from Tapes directory
				strcpy(TmpPath, m_UserDataPath);
				strcat(TmpPath, "tapes/");
				strcat(TmpPath, FileName);
				FILE *fd = fopen(TmpPath, "rb");
				if (fd != NULL)
				{
					cont = true;
					FileName = TmpPath;
					fclose(fd);
				}
			}
		}
		else if (csw)
		{
			// Try getting it from Tapes directory
			strcpy(TmpPath, m_UserDataPath);
			strcat(TmpPath, "tapes/");
			strcat(TmpPath, FileName);
			FILE *fd = fopen(TmpPath, "rb");
			if (fd != NULL)
			{
				cont = true;
				FileName = TmpPath;
				fclose(fd);
			}
		}
		else
		{
			// Try getting it from DiscIms directory
			strcpy(TmpPath, m_UserDataPath);
			strcat(TmpPath, "discims/");
			strcat(TmpPath, FileName);
			FILE *fd = fopen(TmpPath, "rb");
			if (fd != NULL)
			{
				cont = true;
				FileName = TmpPath;
				fclose(fd);
			}
		}

		if (!cont)
		{
			char errstr[200];
			sprintf(errstr,"Cannot find file:\n  %s", FileName);
			MessageBox(m_hWnd,errstr,WindowTitle,MB_OK|MB_ICONERROR);
			cont = false;
		}
	}

	if (cont)
	{
		PathCanonicalize(CmdLineFile, TmpPath);
	}
	else
	{
		CmdLineFile[0] = 0;
	}
#endif
}

/*****************************************************************************/
// Handle a file name passed on command line
	
void BeebWin::HandleCommandLineFile(int drive, char *CmdLineFile)
{
#if 0
	bool ssd = false;
	bool dsd = false;
	bool adfs = false;
	bool cont = false;
	char *FileName = NULL;
	bool uef = false;
	bool csw = false;
	bool img = false;
	
	if (CmdLineFile[0] != 0)
	{
		FileName = CmdLineFile;

		// Work out which type of files it is
		char *ext = strrchr(FileName, '.');
		if (ext != NULL)
		{
			cont = true;
			if (_stricmp(ext+1, "ssd") == 0)
				ssd = true;
			else if (_stricmp(ext+1, "dsd") == 0)
				dsd = true;
			else if (_stricmp(ext+1, "adl") == 0)
				adfs = true;
			else if (_stricmp(ext+1, "adf") == 0)
				adfs = true;
			else if (_stricmp(ext+1, "uef") == 0)
				uef = true;
			else if (_stricmp(ext+1, "csw") == 0)
				csw = true;
			else if (_stricmp(ext+1, "img") == 0)
				img = true;
			else
			{
				char errstr[200];
				sprintf(errstr,"Unrecognised file type:\n  %s", FileName);
				MessageBox(m_hWnd,errstr,WindowTitle,MB_OK|MB_ICONERROR);
				cont = false;
			}
		}
	}

	if (cont)
	{
		cont = false;

		FILE *fd = fopen(FileName, "rb");
		if (fd != NULL)
		{
			cont = true;
			fclose(fd);
		}

		if (!cont)
		{
			char errstr[200];
			sprintf(errstr,"Cannot find file:\n  %s", FileName);
			MessageBox(m_hWnd,errstr,WindowTitle,MB_OK|MB_ICONERROR);
		}
	}

	if (cont)
	{
		if (uef)
		{
			// Determine if file is a tape or a state file
			bool stateFile = false;
			FILE *fd = fopen(FileName, "rb");
			if (fd != NULL)
			{
				char buf[14];
				fread(buf,14,1,fd);
				if (strcmp(buf,"UEF File!")==0 && buf[12]==0x6c && buf[13]==0x04)
				{
					stateFile = true;
				}
				fclose(fd);
			}

			if (stateFile)
				LoadUEFState(FileName);
			else
				LoadUEF(FileName);

			cont = false;
		}
		else if (csw)
		{
			LoadCSW(FileName);
			cont = false;
		}
	}

	if (cont)
	{
		if (MachineType!=3)
		{
			if (dsd)
			{
				if (NativeFDC)
					LoadSimpleDSDiscImage(FileName, drive, 80);
				else
					Load1770DiscImage(FileName,drive,1,m_hMenu); // 1 = dsd
			}
			if (ssd)
			{
				if (NativeFDC)
					LoadSimpleDiscImage(FileName, drive, 0, 80);
				else
					Load1770DiscImage(FileName,drive,0,m_hMenu); // 0 = ssd
			}
			if (adfs)
			{
				if (!NativeFDC)
					Load1770DiscImage(FileName,drive,2,m_hMenu); // 2 = adfs
				else
					cont = false;  // cannot load adfs with native DFS
			}
			if (img)
			{
				if (NativeFDC)
					LoadSimpleDiscImage(FileName, drive, 0, 80); // Treat like an ssd
				else
					Load1770DiscImage(FileName,drive,3,m_hMenu); // 0 = ssd
			}
		}
		else if (MachineType==3)
		{
			if (dsd)
				Load1770DiscImage(FileName,drive,1,m_hMenu); // 0 = ssd
			if (ssd)
				Load1770DiscImage(FileName,drive,0,m_hMenu); // 1 = dsd
			if (adfs)
				Load1770DiscImage(FileName,drive,2,m_hMenu); // 2 = adfs
			if (img)
				Load1770DiscImage(FileName,drive,3,m_hMenu); // 3 = img
		}
	}

	if (cont && !m_NoAutoBoot && drive == 0)
	{
		// Do a shift + break
		ResetBeebSystem(MachineType,TubeEnabled,0);
		BeebKeyDown(0, 0); // Shift key
		m_ShiftBooted = true;
	}
#endif
}

/****************************************************************************/
bool BeebWin::RebootSystem(void)
{
#if 0
	HANDLE hToken;
	TOKEN_PRIVILEGES tkp; 
	
	if (!OpenProcessToken(GetCurrentProcess(), 
						  TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) 
		return( FALSE ); 
	
	LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, 
						 &tkp.Privileges[0].Luid); 
	
	tkp.PrivilegeCount = 1;  // one privilege to set    
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 
	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, 
						  (PTOKEN_PRIVILEGES)NULL, 0); 
	if (GetLastError() != ERROR_SUCCESS) 
		return FALSE; 
	
	if (!ExitWindowsEx(EWX_REBOOT | EWX_FORCE, 
					   SHTDN_REASON_MAJOR_OPERATINGSYSTEM |
					   SHTDN_REASON_MINOR_RECONFIG |
					   SHTDN_REASON_FLAG_PLANNED)) 
		return FALSE; 
	
#endif
	return TRUE;
}

/****************************************************************************/
bool BeebWin::CheckUserDataPath()
{
    bool success = true;
#if 0
	bool copy_user_files = false;
	bool store_user_data_path = false;
	char path[_MAX_PATH];
	char errstr[500];
	DWORD att;

	// Change all '/' to '\'
	for (unsigned int i = 0; i < strlen(m_UserDataPath); ++i)
		if (m_UserDataPath[i] == '/')
			m_UserDataPath[i] = '\\';

	// Check that the folder exists
	att = GetFileAttributes(m_UserDataPath);
	if (att == INVALID_FILE_ATTRIBUTES || !(att & FILE_ATTRIBUTE_DIRECTORY))
	{
		sprintf(errstr, "BeebEm data folder does not exist:\n  %s\n\nCreate the folder?",
				m_UserDataPath);
		if (MessageBox(m_hWnd, errstr, WindowTitle, MB_YESNO|MB_ICONQUESTION) != IDYES)
		{
			// Use data dir installed with BeebEm
			strcpy(m_UserDataPath, m_AppPath);
			strcat(m_UserDataPath, "UserData\\");

			store_user_data_path = true;
		}
		else
		{
			// Create the folder
			strcpy(path, m_UserDataPath);

			char *s = strchr(path, '\\');
			while (s != NULL)
			{
				*s = 0;
				CreateDirectory(path, NULL);
				*s = '\\';
				s = strchr(s+1, '\\');
			}

			att = GetFileAttributes(m_UserDataPath);
			if (att == INVALID_FILE_ATTRIBUTES ||
				!(att & FILE_ATTRIBUTE_DIRECTORY))
			{
				sprintf(errstr, "Failed to create BeebEm data folder:\n  %s", m_UserDataPath);
				MessageBox(m_hWnd, errstr, WindowTitle, MB_OK|MB_ICONERROR);
				success = false;
			}
			else
			{
				copy_user_files = true;
			}
		}
	}
	else
	{
		// Check that essential files are in the user data folder
		sprintf(path, "%sBeebFile", m_UserDataPath);
		att = GetFileAttributes(path);
		if (att == INVALID_FILE_ATTRIBUTES || !(att & FILE_ATTRIBUTE_DIRECTORY))
			copy_user_files = true;
		if (!copy_user_files)
		{
			sprintf(path, "%sBeebState", m_UserDataPath);
			att = GetFileAttributes(path);
			if (att == INVALID_FILE_ATTRIBUTES || !(att & FILE_ATTRIBUTE_DIRECTORY))
				copy_user_files = true;
		}
		if (!copy_user_files)
		{
			sprintf(path, "%sEconet.cfg", m_UserDataPath);
			att = GetFileAttributes(path);
			if (att == INVALID_FILE_ATTRIBUTES)
				copy_user_files = true;
		}
		if (!copy_user_files)
		{
			sprintf(path, "%sAUNMap", m_UserDataPath);
			att = GetFileAttributes(path);
			if (att == INVALID_FILE_ATTRIBUTES)
				copy_user_files = true;
		}
#ifdef SPEECH_ENABLED
		if (!copy_user_files)
		{
			sprintf(path, "%sPhroms.cfg", m_UserDataPath);
			att = GetFileAttributes(path);
			if (att == INVALID_FILE_ATTRIBUTES)
				copy_user_files = true;
		}
#endif
		if (!copy_user_files)
		{
			if (strcmp(RomFile, "Roms.cfg") == 0)
			{
				sprintf(path, "%sRoms.cfg", m_UserDataPath);
				att = GetFileAttributes(path);
				if (att == INVALID_FILE_ATTRIBUTES)
					copy_user_files = true;
			}
		}

		if (copy_user_files)
		{
			sprintf(errstr, "Essential or new files missing from BeebEm data folder:\n  %s"
					"\n\nCopy essential or new files into folder?", m_UserDataPath);
			if (MessageBox(m_hWnd, errstr, WindowTitle, MB_YESNO|MB_ICONQUESTION) != IDYES)
			{
				success = false;
			}
		}
	}

	if (success)
	{
		// Get fully qualified user data path
		char *f;
		if (GetFullPathName(m_UserDataPath, _MAX_PATH, path, &f) != 0)
			strcpy(m_UserDataPath, path);
	}

	if (success && copy_user_files)
	{
		SHFILEOPSTRUCT fileOp = {0};
		fileOp.hwnd = m_hWnd;
		fileOp.wFunc = FO_COPY;
		fileOp.fFlags = 0;

		strcpy(path, m_AppPath);
		strcat(path, "UserData\\*.*");
		path[strlen(path)+1] = 0; // need double 0
		fileOp.pFrom = path;

		m_UserDataPath[strlen(m_UserDataPath)+1] = 0; // need double 0
		fileOp.pTo = m_UserDataPath;

		if (SHFileOperation(&fileOp) || fileOp.fAnyOperationsAborted)
		{
			sprintf(errstr, "Copy failed.  Manually copy files from:\n  %s"
					"\n\nTo BeebEm data folder:\n  %s", path, m_UserDataPath);
			MessageBox(m_hWnd, errstr, WindowTitle, MB_OK|MB_ICONERROR);
			success = false;
		}
		else
		{
			// Wait for copy dialogs to disappear
			Sleep(2000);
		}
	}

	if (success)
	{
		// Check that roms file exists and create its full path
		if (PathIsRelative(RomFile))
		{
			sprintf(path, "%s%s", m_UserDataPath, RomFile);
			strcpy(RomFile, path);
		}
		att = GetFileAttributes(RomFile);
		if (att == INVALID_FILE_ATTRIBUTES)
		{
			sprintf(errstr, "Cannot open ROMs file:\n  %s", RomFile);
			MessageBox(m_hWnd, errstr, WindowTitle, MB_OK|MB_ICONERROR);
			success = false;
		}
	}

	if (success)
	{
		// Fill out full path of prefs file
		if (PathIsRelative(m_PrefsFile))
		{
			sprintf(path, "%s%s", m_UserDataPath, m_PrefsFile);
			strcpy(m_PrefsFile, path);
		}
	}

	if (success && (copy_user_files || store_user_data_path))
	{
		StoreUserDataPath();
	}
#endif
	return success;
}

void BeebWin::StoreUserDataPath()
{
#if 0
	// Store user data path in registry
	RegCreateKey(HKEY_CURRENT_USER, CFG_REG_KEY);
	RegSetStringValue(HKEY_CURRENT_USER, CFG_REG_KEY, "UserDataFolder", m_UserDataPath);
#endif
}


void BeebWin::SelectUserDataPath()
{
#if 0
	char szDisplayName[MAX_PATH];
	char szPathBackup[MAX_PATH];
	BROWSEINFO bi;

	memset(&bi, 0, sizeof(bi));
	bi.hwndOwner = m_hWnd;
	bi.pszDisplayName = szDisplayName;
	bi.lpszTitle = "Select folder for user data files:";
	bi.ulFlags = BIF_EDITBOX | BIF_NEWDIALOGSTYLE;
	bi.lpfn = BrowseUserDataCallbackProc;
	LPITEMIDLIST idList = SHBrowseForFolder(&bi);
	if (idList != NULL)
	{
		if (SHGetPathFromIDList(idList, szExportUserDataPath) == FALSE)
		{
			MessageBox(m_hWnd, "Invalid folder selected", WindowTitle, MB_OK|MB_ICONWARNING);
		}
		else
		{
			strcpy(szPathBackup, m_UserDataPath);
			strcpy(m_UserDataPath, szExportUserDataPath);
			strcat(m_UserDataPath, "\\");

			// Check folder contents
			if (CheckUserDataPath() == false)
			{
				strcpy(m_UserDataPath, szPathBackup);
			}
			else
			{
				// Store the new path
				StoreUserDataPath();

				// Reset prefs and roms file paths
				strcpy(m_PrefsFile, "Preferences.cfg");
				strcpy(RomFile, "Roms.cfg");
				CheckForLocalPrefs(m_UserDataPath, true);

				// Load and apply prefs
				LoadPreferences();
				ApplyPrefs();
			}
		}
		CoTaskMemFree(idList);
	}
#endif
}

/****************************************************************************/
void BeebWin::HandleTimer()
{
#if 0
	int row,col;
	char delay[10];

	// Do nothing if emulation is not running (e.g. if Window is being moved)
	if ((TotalCycles - m_KbdCmdLastCycles) / 2000 < m_KbdCmdDelay)
	{
		SetTimer(m_hWnd, 1, m_KbdCmdDelay, NULL);
		return;
	}
	m_KbdCmdLastCycles = TotalCycles;
	
	// Release previous key press (except shift/control)
	if (m_KbdCmdPress &&
		m_KbdCmdKey != VK_SHIFT && m_KbdCmdKey != VK_CONTROL)
	{
		TranslateKey(m_KbdCmdKey, 1, row, col);
		m_KbdCmdPress = false;
		SetTimer(m_hWnd, 1, m_KbdCmdDelay, NULL);
	}
	else
	{
		m_KbdCmdPos++;
		if (m_KbdCmd[m_KbdCmdPos] == 0)
		{
			KillTimer(m_hWnd, 1);
		}
		else
		{
			m_KbdCmdPress = true;
			
			switch (m_KbdCmd[m_KbdCmdPos])
			{
			case '\\':
				m_KbdCmdPos++;
				switch (m_KbdCmd[m_KbdCmdPos])
				{
				case '\\': m_KbdCmdKey = VK_OEM_5; break;
				case 'n': m_KbdCmdKey = VK_RETURN; break;
				case 's': m_KbdCmdKey = VK_SHIFT; break;
				case 'S': m_KbdCmdKey = VK_SHIFT; m_KbdCmdPress = false; break;
				case 'c': m_KbdCmdKey = VK_CONTROL; break;
				case 'C': m_KbdCmdKey = VK_CONTROL; m_KbdCmdPress = false; break;
				case 'd':
					m_KbdCmdKey = 0;
					m_KbdCmdPos++;
					delay[0] = m_KbdCmd[m_KbdCmdPos];
					m_KbdCmdPos++;
					delay[1] = m_KbdCmd[m_KbdCmdPos];
					m_KbdCmdPos++;
					delay[2] = m_KbdCmd[m_KbdCmdPos];
					m_KbdCmdPos++;
					delay[3] = m_KbdCmd[m_KbdCmdPos];
					delay[4] = 0;
					m_KbdCmdDelay = atoi(delay);
					break;
				default: m_KbdCmdKey = 0; break;
				}
				break;
				
			case '`': m_KbdCmdKey = VK_OEM_8; break;
			case '-': m_KbdCmdKey = VK_OEM_MINUS; break;
			case '=': m_KbdCmdKey = VK_OEM_PLUS; break;
			case '[': m_KbdCmdKey = VK_OEM_4; break;
			case ']': m_KbdCmdKey = VK_OEM_6; break;
			case ';': m_KbdCmdKey = VK_OEM_1; break;
			case '\'': m_KbdCmdKey = VK_OEM_3; break;
			case '#': m_KbdCmdKey = VK_OEM_7; break;
			case ',': m_KbdCmdKey = VK_OEM_COMMA; break;
			case '.': m_KbdCmdKey = VK_OEM_PERIOD; break;
			case '/': m_KbdCmdKey = VK_OEM_2; break;
			default: m_KbdCmdKey = m_KbdCmd[m_KbdCmdPos]; break;
			}
			
			if (m_KbdCmdKey != 0)
			{
				if (m_KbdCmdPress)
					TranslateKey(m_KbdCmdKey, 0, row, col);
				else
					TranslateKey(m_KbdCmdKey, 1, row, col);
			}

			SetTimer(m_hWnd, 1, m_KbdCmdDelay, NULL);
		}
	}
#endif
}

