/****************************************************************
 BeebEm - BBC Micro and Master 128 Emulator
 Copyright (C) 1994  David Alan Gilbert
 Copyright (C) 1994  Nigel Magnay
 Copyright (C) 1997  Mike Wyatt
 Copyright (C) 2001  Richard Gellman
 Copyright (C) 2008  Rich Talbot-Watkins
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 Boston, MA  02110-1301, USA.
 ****************************************************************/

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "BeebFrame.h"
#include "FrameQueue.h"
#include "VideoState.h"

#include "TeletextVideo.h"

//#define BEEB_DOTIME

#include "6502core.h"
#include "BeebMemory.h"
/*
#include "beebwin.h"
 */
#if 0
#include "main.h"
#endif
#include "sysvia.h"
#include "video.h"
#if 0
#include "uefstate.h"
#include "beebsound.h"
#endif
#include "debug.h"
#if 0
#include "teletext.h"
#endif
#ifdef BEEB_DOTIME
#include <sys/times.h>
#endif


CBeebFrame *currentFrame = NULL;
EMVIDEOLINK *VideoLink = NULL;

unsigned char VideoULA_ControlReg = 0x9c;
unsigned char VideoULA_Palette[16];

unsigned char CRTCControlReg = 0;
unsigned char CRTC_HorizontalTotal = 127;     /* R0 */
unsigned char CRTC_HorizontalDisplayed = 80;  /* R1 */
unsigned char CRTC_HorizontalSyncPos = 98;    /* R2 */
unsigned char CRTC_SyncWidth = 0x28;          /* R3 - top 4 bits are Vertical (in scan lines) and bottom 4 are horizontal in characters */
unsigned char CRTC_VerticalTotal = 38;        /* R4 */
unsigned char CRTC_VerticalTotalAdjust = 0;   /* R5 */
unsigned char CRTC_VerticalDisplayed = 32;    /* R6 */
unsigned char CRTC_VerticalSyncPos = 34;      /* R7 */
unsigned char CRTC_InterlaceAndDelay = 0;     /* R8 - 0,1 are interlace modes, 4,5 display blanking delay, 6,7 cursor blanking delay */
unsigned char CRTC_ScanLinesPerChar = 7;      /* R9 */
unsigned char CRTC_CursorStart = 0;           /* R10 */
unsigned char CRTC_CursorEnd = 0;             /* R11 */
unsigned char CRTC_ScreenStartHigh = 6;       /* R12 */
unsigned char CRTC_ScreenStartLow = 0;        /* R13 */
unsigned char CRTC_CursorPosHigh = 0;         /* R14 */
unsigned char CRTC_CursorPosLow = 0;          /* R15 */
unsigned char CRTC_LightPenHigh = 0;          /* R16 */
unsigned char CRTC_LightPenLow = 0;           /* R17 */

unsigned int ActualScreenWidth = 640;
int InitialOffset = 0;
int ScreenAdjust = 0; // Mode 7 Defaults.
int VScreenAdjust = 0;
int VStart, HStart;
unsigned char HSyncModifier = 9;
unsigned char TeletextEnabled = 0;
int THalfMode = 0; // 1 if to use half-mode (TeletextStyle=1 all the time)
int CurY = -1;
FILE *crtclog;

int ova, ovn; // mem ptr buffers



VideoStateT VideoState;

CycleCountT VideoTriggerCount = 9999; /* Number of cycles before next scanline service */

unsigned char CursorOnFields,CursorOffFields;
int CursorFieldCount=32;
unsigned char CursorOnState=1;

/* If 1 then refresh on every display, else refresh every n'th display */
int Video_RefreshFrequency=1;
/* The number of the current frame - starts at Video_RefreshFrequency - at 0 actually refresh */
int FrameNum = 0;


void VideoAddCursor(void);
void AdjustVideo();
void VideoAddLEDs(void);


extern int FastTable_Valid;
void LowLevelDoScanLine();


/*-------------------------------------------------------------------------------------------------------------*/
#define BEEB_DOTIME_SAMPLESIZE 50

void VideoStartOfFrame(void)
{
    static int InterlaceFrame=0;
    static int frames = 0;
    int CurStart;
    int IL_Multiplier;
    
#ifdef BEEB_DOTIME
    static int Have_GotTime=0;
    //    static struct tms previous,now;
    static clock_t previous;
    clock_t now;
    static int Time_FrameCount=0;
    
    double frametime;
    static CycleCountT OldCycles=0;
    
    //    int CurStart;
    
    frames++;
    
    if(!Have_GotTime)
    {
        previous = clock();
        //        times(&previous);
        Time_FrameCount=-1;
        Have_GotTime=1;
    }
    
    if(Time_FrameCount == (BEEB_DOTIME_SAMPLESIZE-1))
    {
        now = clock();
        //        times(&now);
        //        frametime = now.tms_utime-previous.tms_utime;
        frametime = (double)(now - previous);
        frametime /= (double)CLOCKS_PER_SEC;
        
        //        frametime /= (double)BEEB_DOTIME_SAMPLESIZE;
        fprintf(stderr, "Time: %f Frame: %d Frametime: %fs fps=%f Total cycles = %d Cycles in last unit = %d\n", (double)now / CLOCKS_PER_SEC, frames, frametime,1/frametime, TotalCycles, TotalCycles-OldCycles);
        //        std::cerr << "Frametime: " << frametime << "s fps=" << (1/frametime) << "Total cycles=" << TotalCycles << "Cycles in last unit=" << (TotalCycles-OldCycles) << "\n";
        OldCycles = TotalCycles;
        previous = now;
        Time_FrameCount = 0;
        
    }
    else
    {
        Time_FrameCount++;
    }
#endif
    
    /* FrameNum is determined by the window handler */
    if(VideoState.IsNewTVFrame)	// RTW - only calibrate timing once per frame
    {
        VideoState.IsNewTVFrame = 0;

        /* If FrameNum hits 0 we actually refresh */
        if (FrameNum-- == 0)
        {
            FrameNum = Video_RefreshFrequency - 1;
        }
        
        if(currentFrame && VideoLink)
        {
            /* Queue frame, something else will delete it */
            pthread_mutex_lock(&VideoLink->queueMutex);
            VideoLink->frameQueue->push(currentFrame);
            pthread_mutex_unlock(&VideoLink->queueMutex);
        }
        
        currentFrame = new CBeebFrame();
        
        CursorFieldCount--;
        Mode7FlashTrigger--;
        InterlaceFrame ^= 1;
    }
    
    // Cursor update for blink. I thought I'd put it here, as this is where the mode 7 flash field thingy is too
    // - Richard Gellman
    if (CursorFieldCount < 0)
    {
        CurStart = CRTC_CursorStart & 0x60;
        
        // 0 is cursor displays, but does not blink
        // 32 is no cursor
        // 64 is 1/16 fast blink
        // 96 is 1/32 slow blink
        if (CurStart==0)
        {
            CursorFieldCount = CursorOnFields;
            CursorOnState=1;
        }
        
        if (CurStart==32)
        {
            CursorFieldCount = CursorOffFields;
            CursorOnState=0;
        }
        if (CurStart==64) { CursorFieldCount=8; CursorOnState^=1; }
        if (CurStart==96) { CursorFieldCount=16; CursorOnState^=1; }
    }
    
    // RTW - The meaning of CharLine has changed: -1 no longer means that we are in the vertical
    // total adjust period, and this is no longer handled as if it were at the beginning of a new CRTC cycle.
    // Hence, here we always set CharLine to 0.
    VideoState.CharLine=0;
    VideoState.InCharLineUp=0;
    
    VideoState.IsTeletext=(VideoULA_ControlReg &2)>0;
    if(!VideoState.IsTeletext)
    {
        VideoState.Addr=VideoState.StartAddr=CRTC_ScreenStartLow+(CRTC_ScreenStartHigh<<8);
    }
    else
    {
        int tmphigh=CRTC_ScreenStartHigh;
        /* undo wrangling of start address - I don't understand why this should be - see p.372 of AUG for this info */
        tmphigh ^= 0x20;
        tmphigh += 0x74;
        tmphigh &= 255;
        VideoState.Addr=VideoState.StartAddr=CRTC_ScreenStartLow+(tmphigh<<8);
        
        // O aye. this is the mode 7 flash section is it? Modified for corrected flash settings - Richard Gellman
        if (Mode7FlashTrigger < 0)
        {
            Mode7FlashTrigger = (Mode7FlashOn) ? MODE7OFFFIELDS : MODE7ONFIELDS;
            Mode7FlashOn ^= 1; /* toggle flash state */
        }
    };
    
    IL_Multiplier=(CRTC_InterlaceAndDelay&1)?2:1;
    if (InterlaceFrame)
    {
        IncTrigger((IL_Multiplier*(CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2)),VideoTriggerCount); /* Number of 2MHz cycles until another scanline needs doing */
    }
    else
    {
        IncTrigger(((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2)),VideoTriggerCount); /* Number of 2MHz cycles until another scanline needs doing */
    };
    
#if 0
    TeleTextPoll();
#endif
} /* VideoStartOfFrame */


/*-------------------------------------------------------------------------------------------------------------*/
/* Actually does the work of decoding beeb memory and plotting the line to X */


void RedoMPTR(CBeebMemory *memory)
{
    if (VideoState.IsTeletext)
    {
        VideoState.DataPtr = memory->PtrWithWrapMo7(ova, ovn);
    }
    
    if (!VideoState.IsTeletext)
    {
        VideoState.DataPtr = memory->PtrWithWrap(ova, ovn);
    }
    
    //FastTable_Valid=0;
}

/*-------------------------------------------------------------------------------------------------------------*/
void VideoDoScanLine(CBeebMemory *memory)
{
    /* cerr << "CharLine=" << VideoState.CharLine << " InCharLineUp=" << VideoState.InCharLineUp << "\n"; */
    if (VideoState.IsTeletext)
    {
        TeletextVideoDoScanLine(memory);
    }
    else
    {
        /* Non teletext. */
        
        // Handle VSync
        // RTW - this was moved to the top so that we can correctly set R7=0,
        // i.e. we can catch it before the line counters are incremented
        if (VideoState.VSyncState)
        {
            if (!(--VideoState.VSyncState))
            {
                SysVIATriggerCA1Int(0);
            };
        }
        
        if ((VideoState.VSyncState==0) && (VideoState.CharLine==CRTC_VerticalSyncPos) && (VideoState.InCharLineUp == 0))
        {
            // Nothing displayed?
            if (VideoState.FirstPixmapLine < 0)
            {
                VideoState.FirstPixmapLine = 0;
            }
            
            VideoState.PreviousFirstPixmapLine = VideoState.FirstPixmapLine;
            VideoState.FirstPixmapLine = -1;
            VideoState.PreviousLastPixmapLine = VideoState.LastPixmapLine;
            VideoState.LastPixmapLine = 0;
            VideoState.PixmapLine = 0;
            VideoState.IsNewTVFrame = 1;
            
            SysVIATriggerCA1Int(1);
            VideoState.VSyncState = (CRTC_SyncWidth >> 4);
        }
        
        /* Clear the scan line */
        if(!FrameNum)
        {
            currentFrame->ClearScanLine(VideoState.PixmapLine * 2);
            currentFrame->ClearScanLine(VideoState.PixmapLine * 2 + 1);
        }

        // RTW - changed so we are even able to plot vertical total adjust region if CRTC_VerticalDisplayed is high enough
        if (VideoState.CharLine < CRTC_VerticalDisplayed)
        {
            // Visible char line, record first line
            if (VideoState.FirstPixmapLine==-1)
            {
                VideoState.FirstPixmapLine=VideoState.PixmapLine;
            }
            
            // Always record the last line
            VideoState.LastPixmapLine=VideoState.PixmapLine;
            
            /* If first row of character then get the data pointer from memory */
            if (VideoState.InCharLineUp==0)
            {
                ova = VideoState.Addr * 8;
                ovn = CRTC_HorizontalDisplayed * 8;
                VideoState.DataPtr = memory->PtrWithWrap(VideoState.Addr * 8, CRTC_HorizontalDisplayed * 8);
                VideoState.Addr += CRTC_HorizontalDisplayed;
            }
            
            if ((VideoState.InCharLineUp < 8) && ((CRTC_InterlaceAndDelay & 0x30) != 48))
            {
                if (!FrameNum)
                {
                    LowLevelDoScanLine();
                }
            }
        }
        
        // See if we are at the cursor line
        if (CurY == -1 && VideoState.Addr > (CRTC_CursorPosLow+(CRTC_CursorPosHigh<<8)))
            CurY = VideoState.PixmapLine;
        
        // Screen line increment and wraparound
        if (++VideoState.PixmapLine == MAX_VIDEO_SCAN_LINES)
            VideoState.PixmapLine = 0;
        
        /* Move onto next physical scanline as far as timing is concerned */
        VideoState.InCharLineUp += 1;
        
        // RTW - check whether we have reached a new character row.
        // if CharLine>CRTC_VerticalTotal, we are in the vertical total adjust region so we don't wrap to a new row.
        if (VideoState.CharLine <= CRTC_VerticalTotal && VideoState.InCharLineUp > CRTC_ScanLinesPerChar)
        {
            VideoState.CharLine++;
            VideoState.InCharLineUp = 0;
        }
        
        // RTW - neater way of detecting the end of the PAL frame, which doesn't require making a special case
        // of the vertical total adjust period.
        if (VideoState.CharLine > CRTC_VerticalTotal && VideoState.InCharLineUp >= CRTC_VerticalTotalAdjust)
        {
            VScreenAdjust=0;
            if (!FrameNum && VideoState.IsNewTVFrame)
            {
                VideoAddCursor();
                VideoAddLEDs();
                CurY=-1;
                int n = VideoState.PreviousLastPixmapLine - VideoState.PreviousFirstPixmapLine + 1;
                if (n < 0)
                {
                    n += MAX_VIDEO_SCAN_LINES;
                }
                
                int startLine = 32;
                if (n > 248 && VideoState.PreviousFirstPixmapLine >= 40)
                {
                    // RTW -
                    // This is a little hack which ensures that a fullscreen mode with *TV255 will always
                    // fit unclipped in the window in Modes 0-6
                    startLine = 40;
                }

            }
            VideoStartOfFrame();
            AdjustVideo();
        }
        else
        {
            IncTrigger((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2),VideoTriggerCount);
        }
    } /* Teletext if */
    
} /* VideoDoScanLine */

/*-------------------------------------------------------------------------------------------------------------*/
void AdjustVideo()
{
    ActualScreenWidth = CRTC_HorizontalDisplayed * HSyncModifier;
    if(ActualScreenWidth > CBeebFrame::kFrameWidth)
    {
        ActualScreenWidth = CBeebFrame::kFrameWidth;
    }
    
    if (ActualScreenWidth < 640)
    {
        ActualScreenWidth = 640;
    }
    
    InitialOffset=0-(((CRTC_HorizontalTotal+1)/2)-((HSyncModifier==8)?40:20));
    HStart=InitialOffset+((CRTC_HorizontalTotal+1)-(CRTC_HorizontalSyncPos+(CRTC_SyncWidth&15)));
    HStart+=(HSyncModifier==8)?2:1;
    
    if (TeletextEnabled)
    {
        HStart+=2;
    }
    
    if (HStart<0)
    {
        HStart=0;
    }
    
    ScreenAdjust = (HStart * HSyncModifier) + ((VScreenAdjust > 0) ? (VScreenAdjust * 800) : 0);
}
/*-------------------------------------------------------------------------------------------------------------*/
void VideoInit(CBeebMemory *memory)
{
    //  char *environptr;
    VideoStartOfFrame();
    ova=0x3000; ovn=640;
    VideoState.DataPtr = memory->PtrWithWrap(0x3000,640);
    SetTrigger(99,VideoTriggerCount); /* Give time for OS to set mode up before doing anything silly */
    FastTable_Valid=0;
#if 0
    BuildMode7Font();
    

#ifndef WIN32
    environptr=getenv("BeebVideoRefreshFreq");
    if (environptr!=NULL) Video_RefreshFrequency=atoi(environptr);
    if (Video_RefreshFrequency<1) Video_RefreshFrequency=1;
#endif
#endif
    
    Video_RefreshFrequency = 1;
    
    FrameNum=Video_RefreshFrequency;
    VideoState.PixmapLine=0;
    VideoState.FirstPixmapLine=-1;
    VideoState.PreviousFirstPixmapLine=0;
    VideoState.LastPixmapLine=0;
    VideoState.PreviousLastPixmapLine=256;
    VideoState.IsNewTVFrame=0;
    CurY=-1;
    //AdjustVideo();
    //  crtclog=fopen("/crtc.log","wb");
} /* VideoInit */


/*-------------------------------------------------------------------------------------------------------------*/
void CRTCWrite(int Address, int Value)
{
    Value &= 0xff;
    if (Address & 1)
    {
        //	if (CRTCControlReg<14) { fputc(CRTCControlReg,crtclog); fputc(Value,crtclog); }
        //	if (CRTCControlReg<14) {
        //		fprintf(crtclog,"%d (%02X) Written to register %d from %04X\n",Value,Value,CRTCControlReg,ProgramCounter);
        //	}
        
        if (DebugEnabled && CRTCControlReg < 14)
        {
            char info[200];
            sprintf(info, "CRTC: Write register %X value %02X", (int)CRTCControlReg, Value);
            DebugDisplayTrace(DEBUG_VIDEO, true, info);
        }
        
        switch (CRTCControlReg)
        {
            case 0:
                CRTC_HorizontalTotal = Value;
                InitialOffset = 0 - (((CRTC_HorizontalTotal + 1) / 2) - ((HSyncModifier == 8) ? 40 : 20));
                AdjustVideo();
                break;
                
            case 1:
                CRTC_HorizontalDisplayed = Value;
                if (CRTC_HorizontalDisplayed > 127)
                {
                    CRTC_HorizontalDisplayed = 127;
                }
                FastTable_Valid = 0;
                AdjustVideo();
                break;
                
            case 2:
                CRTC_HorizontalSyncPos = Value;
                AdjustVideo();
                break;
                
            case 3:
                CRTC_SyncWidth=Value;
                AdjustVideo();
                //		fprintf(crtclog,"V Sync width: %d\n",(Value>>4)&15);
                break;
                
            case 4:
                CRTC_VerticalTotal=Value;
                AdjustVideo();
                break;
                
            case 5:
                CRTC_VerticalTotalAdjust=Value & 0x1f;  // 5 bit register
                //		fprintf(crtclog,"Vertical Total Adjust: %d\n",Value);
                AdjustVideo();
                break;
                
            case 6:
                CRTC_VerticalDisplayed = Value;
                AdjustVideo();
                break;
                
            case 7:
                CRTC_VerticalSyncPos = (Value & 0x7f);
                AdjustVideo();
                break;
                
            case 8:
                CRTC_InterlaceAndDelay = Value;
                break;
                
            case 9:
                CRTC_ScanLinesPerChar = Value;
                break;
                
            case 10:
                CRTC_CursorStart = Value;
                break;
                
            case 11:
                CRTC_CursorEnd = Value;
                break;
                
            case 12:
                CRTC_ScreenStartHigh = Value;
                //		fprintf(crtclog,"Screen now at &%02x%02x\n",Value,CRTC_ScreenStartLow);
                break;
                
            case 13:
                CRTC_ScreenStartLow = Value;
                //		fprintf(crtclog,"Screen now at &%02x%02x\n",CRTC_ScreenStartHigh,Value);
                break;
                
            case 14:
                CRTC_CursorPosHigh = Value & 0x3f; /* Cursor high only has 6 bits */
                break;
                
            case 15:
                CRTC_CursorPosLow = Value & 0xff;
                break;
                
            default: /* In case the user wrote a duff control register value */
                break;
        }; /* CRTCWrite switch */
        /* cerr << "CRTCWrite RegNum=" << int(CRTCControlReg) << " Value=" << Value << "\n"; */
    }
    else
    {
        CRTCControlReg = Value & 0x1f;
    }
    
} /* CRTCWrite */

/*-------------------------------------------------------------------------------------------------------------*/
int CRTCRead(int Address)
{
    if (Address & 1)
    {
        switch (CRTCControlReg)
        {
            case 14:
                return(CRTC_CursorPosHigh);
            case 15:
                return(CRTC_CursorPosLow);
            case 16:
                return(CRTC_LightPenHigh); /* Perhaps tie to mouse pointer ? */
            case 17:
                return(CRTC_LightPenLow);
            default:
                break;
        } /* CRTC Read switch */
    }
    else
    {
        return(0); /* Rockwell part has bits 5,6,7 used - bit 6 is set when LPEN is received, bit 5 when in vertical retrace */
    }
    return(0);	// Keeep MSVC happy $NRM
}; /* CRTCRead */

/*-------------------------------------------------------------------------------------------------------------*/
void VideoULAWrite(int Address, int Value)
{
    int oldValue;
    if (Address & 1)
    {
        VideoULA_Palette[(Value & 0xf0) >> 4] = (Value & 0xf) ^ 7;
        FastTable_Valid = 0;
        /* cerr << "Palette reg " << ((Value & 0xf0)>>4) << " now has value " << ((Value & 0xf) ^ 7) << "\n"; */
        //fprintf(crtclog,"Pallette written to at line %d\n",VideoState.PixmapLine);
    }
    else
    {
        oldValue = VideoULA_ControlReg;
        VideoULA_ControlReg = Value;
        FastTable_Valid=0; /* Could be more selective and only do it if no.of.cols bit changes */
        /* cerr << "VidULA Ctrl reg write " << hex << Value << "\n"; */
        // Adjust HSyncModifier
        if (VideoULA_ControlReg & 16) HSyncModifier=8; else HSyncModifier=16;
        if (VideoULA_ControlReg & 2) HSyncModifier=12;
        // number of pixels per CRTC character (on our screen)
        if (Value & 2) TeletextEnabled=1; else TeletextEnabled=0;
        if ((Value&2)^(oldValue&2)) { ScreenAdjust=0; }
        AdjustVideo();
    };
}; /* VidULAWrite */

/*-------------------------------------------------------------------------------------------------------------*/
int VideoULARead(int Address)
{
    return(Address); /* Read not defined from Video ULA */
}; /* VidULARead */

/*-------------------------------------------------------------------------------------------------------------*/
void VideoAddCursor(void)
{
    static int CurSizes[] = { 2,1,0,0,4,2,0,4 };
    int ScrAddr,CurAddr,RelAddr;
    int CurX;
    int CurSize;
    int CurStart, CurEnd;
    
    /* Check if cursor has been hidden */
    if ((VideoULA_ControlReg & 0xe0) == 0 || (CRTC_CursorStart & 0x60) == 0x20)
        return;
    
    /* Use clock bit and cursor bits to work out size */
    if (VideoULA_ControlReg & 0x80)
    {
        CurSize = CurSizes[(VideoULA_ControlReg & 0x70)>>4] * 8;
    }
    else
    {
        CurSize = 2 * 8; /* Mode 7 */
    }
    
    if (VideoState.IsTeletext)
    {
        ScrAddr = CRTC_ScreenStartLow + (((CRTC_ScreenStartHigh ^ 0x20) + 0x74 & 0xff) << 8);
        CurAddr = CRTC_CursorPosLow + (((CRTC_CursorPosHigh ^ 0x20) + 0x74 & 0xff) << 8);
        
        CurStart = (CRTC_CursorStart & 0x1f) / 2;
        CurEnd = CRTC_CursorEnd;
        CurSize -= 4;
    }
    else
    {
        ScrAddr = CRTC_ScreenStartLow + (CRTC_ScreenStartHigh << 8);
        CurAddr = CRTC_CursorPosLow + (CRTC_CursorPosHigh << 8);
        
        CurStart = CRTC_CursorStart & 0x1f;
        CurEnd = CRTC_CursorEnd;
    }
    
    RelAddr=CurAddr-ScrAddr;
    if (RelAddr < 0 || CRTC_HorizontalDisplayed == 0)
    {
        return;
    }
    
    /* Work out char positions */
    CurX = RelAddr % CRTC_HorizontalDisplayed;
    
    /* Convert to pixel positions */
    if (VideoState.IsTeletext)
    {
        CurX = CurX * 12;
        CurY = (RelAddr / CRTC_HorizontalDisplayed) * 20 + 9;
    }
    else
    {
        CurX = CurX * HSyncModifier;
    }
    
    /* Limit cursor size */ // This should be 11, not 9 - Richard Gellman
    if (CurEnd > 11)
    {
        CurEnd = 11;
    }
    
    if (CurX + CurSize >= 640)
    {
        CurSize = 640 - CurX;
    }
    
    // Cursor delay
    CurX+=((CRTC_InterlaceAndDelay&192)>>6)*HSyncModifier;
    if (VideoState.IsTeletext)
    {
        CurX-=2*HSyncModifier;
    }
    
    if (CurSize > 0)
    {
        for (int y = CurStart; y <= CurEnd && y <= CRTC_ScanLinesPerChar && CurY + y < 500; ++y)
        {
            if (CurY + y >= 0)
            {
                if (CursorOnState)
                {
                    if(VideoState.IsTeletext)
                    {
                        currentFrame->doInvHorizLine(7, (CurY + y), CurX, CurSize);
                    }
                    else
                    {
                        currentFrame->doInvHorizLine(7, (CurY + y)*2, CurX, CurSize);
                        currentFrame->doInvHorizLine(7, (CurY + y)*2+1, CurX, CurSize);
                    }
                }
            }
        }
    }
}

void VideoAddLEDs(void)
{
#if 0
    // now add some keyboard leds
    if (LEDs.ShowKB)
    {
        if (MachineType==3) mainWin->doLED(4,TRUE); else mainWin->doLED(4,LEDs.Motor);
        mainWin->doLED(14,LEDs.CapsLock);
        mainWin->doLED(24,LEDs.ShiftLock);
    }
    if (LEDs.ShowDisc)  {
        int adj = (TeletextEnabled ? 86 : 0);
        mainWin->doLED(578-adj,LEDs.HDisc[0]);
        mainWin->doLED(588-adj,LEDs.HDisc[1]);
        mainWin->doLED(598-adj,LEDs.HDisc[2]);
        mainWin->doLED(608-adj,LEDs.HDisc[3]);
        mainWin->doLED(618-adj,LEDs.Disc0);
        mainWin->doLED(628-adj,LEDs.Disc1);
    }
#endif
}

